#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Data model'''

from __future__ import annotations
import typing
from enum import Enum
from fractions import Fraction
from dataclasses import dataclass
import re
from ttconv.style_properties import StyleProperties
from ttconv.style_properties import StyleProperty
#
# Content Elements
#

class WhiteSpaceHandling(Enum):
  '''Enumerates the strategy for handling white spaces in text nodes.'''
  PRESERVE = "preserve"
  DEFAULT = "default"

@dataclass(frozen=True)
class DiscreteAnimationStep:
  '''Represents a discrete change in the value of a style property over time (see TTML `set` element)
  '''
  style_property: typing.Type[StyleProperty]
  begin: typing.Optional[Fraction]
  end: typing.Optional[Fraction]
  value: typing.Any

  def __post_init__(self):
    if self.style_property is None:
      raise ValueError("style_property cannot be None")

    if self.value is None:
      raise ValueError("value cannot be None")

    if self.style_property not in StyleProperties.ALL:
      raise ValueError("Invalid style property")

    if not self.style_property.validate(self.value):
      raise ValueError("Invalid style property value")

    

class ContentElement:
  '''Abstract base class for all content elements in the model.'''

  def __init__(self, doc=None):

    # space handling

    self._space = WhiteSpaceHandling.DEFAULT

    # language

    self._lang = ""
    
    # owner document

    self._doc = doc

    # hierarchy

    self._first_child = None
    self._last_child = None
    self._parent = None
    self._previous_sibling = None
    self._next_sibling = None

    # styles

    self._styles = {}

    # animation

    self._sets = []

    # layout

    self._region = None

    # timing

    self._begin = None
    self._end = None

    # id

    self._id = None

  def copy_to(self, dest: ContentElement):
    '''Copy all information but children to a different instance.
    '''

    if dest is self:
      return

    dest.set_begin(self.get_begin())
    dest.set_end(self.get_end())
    dest.set_id(self.get_id())
    dest.set_lang(self.get_lang())
    dest.set_space(self.get_space())
    
    for style_prop in self.iter_styles():
      dest.set_style(style_prop, self.get_style(style_prop))

    for anim_step in self.iter_animation_steps():
      dest.add_animation_step(anim_step)

  # document

  def is_attached(self) -> bool:
    '''Returns whether the element belongs to a `ContentDocument`.'''
    return self.get_doc() is not None

  def get_doc(self) -> typing.Optional[ContentDocument]:
    '''Returns the `ContentDocument` to which the element belongs, or `None` otherwise.'''
    return self._doc

  def set_doc(self, doc: ContentDocument):
    '''Attaches the element to `doc`, or detaches it from its current owning
    `ContentDocument` if `doc` is `None`.'''

    if doc is None:

      # detaching

      if self.parent() is not None:
        raise RuntimeError("Element must be removed from parent first")

      self.set_region(None)

    else:

      # attaching

      for e in self.dfs_iterator():
        if e.is_attached():
          raise RuntimeError("Element must be detached first")

    self._doc = doc

    for e in self:
      e.set_doc(doc)

  # hierarchical structure

  def has_children(self) -> bool:
    '''Returns whether the element has children elements.'''
    return self._first_child is not None

  def first_child(self) -> typing.Optional[ContentElement]:
    '''Returns the first child of the element, or `None` if the element has no children.'''
    return self._first_child

  def last_child(self) -> typing.Optional[ContentElement]:
    '''Returns the last child of the element, or `None` if the element has no children.'''
    return self._last_child

  def root(self) -> ContentElement:
    '''Returns the root of the tree to which the element belongs.'''
    root = self

    while root.parent() is not None:
      root = root.parent()

    return root

  def next_sibling(self) ->  typing.Optional[ContentElement]:
    '''Returns the next sibling of the element, or None if none exists.'''
    return self._next_sibling

  def previous_sibling(self) -> typing.Optional[ContentElement]:
    '''Returns the previous sibling of the element, or None if none exists.'''
    return self._previous_sibling

  def parent(self) -> typing.Optional[ContentElement]:
    '''Returns the parent of the element, or None if the element is the root of the hierarchy.'''
    return self._parent

  def remove(self):
    '''Removes the element from the list of children of its parent,
    or does nothing if the element is the root.'''

    if self._parent is None:
      return

    self._parent.remove_child(self)

  def remove_child(self, child: ContentElement):
    '''Remove `child` from the list of children of the element.'''

    if child not in list(self):
      raise ValueError("Element is not a child of this element")

    # pylint: disable=W0212

    if self._first_child is child:
      self._first_child = child._next_sibling

    if self._last_child is child:
      self._last_child = child._previous_sibling

    if child._previous_sibling is not None:
      child._previous_sibling._next_sibling = child._next_sibling

    if child._next_sibling is not None:
      child._next_sibling._previous_sibling = child._previous_sibling

    child._parent = None
    child._next_sibling = None
    child._previous_sibling = None

    # pylint: enable=W0212

  def remove_children(self):
    '''Remove all children of the element.'''

    for c in list(self):
      self.remove_child(c)

  def push_child(self, child: ContentElement):
    '''Adds `child` as a child of the element.'''

    if child.parent() is not None:
      raise RuntimeError("Element already has a parent")

    if child.get_doc() != self.get_doc():
      raise RuntimeError("Element belongs to a different document")

    if child is self:
      raise RuntimeError("Cannot add a root element to its descendents")

    # pylint: disable=W0212

    child._parent = self

    child._previous_sibling = self._last_child
    child._next_sibling = None

    if self._last_child is not None:
      self._last_child._next_sibling = child

    if self._first_child is None:
      self._first_child = child

    self._last_child = child

    # pylint: enable=W0212

  def push_children(self, children: typing.Iterable[ContentElement]):
    '''Adds `children` as children of the element. Unless overridden, this method simply
    calls `push_child` repeatedly. Elements that require children to appear in a certain
    order can override this method.'''

    for c in children:
      self.push_child(c)

  def __iter__(self) -> typing.Iterator[ContentElement]:
    '''Returns an iterator over the children of the element.'''
    child = self._first_child
    while child is not None:
      yield child
      child = child._next_sibling

  def __len__(self) -> int:
    '''Returns the number of children of the element.'''
    count = 0
    child = self._first_child
    while child is not None:
      count += 1
      child = child._next_sibling
    return count

  def __getitem__(self, key: int) -> ContentElement:
    '''Returns the key`th child of the element.'''
    return list(self)[key]

  def dfs_iterator(self) -> typing.Iterator[ContentElement]:
    '''Returns an iterator over all elements in the tree rooted at the element,
    in depth-first search order.'''
    yield self
    for c in self:
      yield from c.dfs_iterator()

  # id property

  def get_id(self) -> str:
    '''Returns the `id` of the element'''
    return self._id

  _XML_ID_REGEXP = re.compile(r'^[a-zA-Z_][\w.-]*$')

  def set_id(self, element_id: str):
    '''Sets the `id` of the element'''
    if not (
        (isinstance(element_id, str) and self._XML_ID_REGEXP.match(element_id)) 
        or element_id is None
      ):
      raise TypeError("Element id must be a valid xml:id string")

    self._id = element_id
  
  # style properties

  def has_style(self, style_prop: typing.Type[StyleProperty]) -> bool:
    '''Returns whether there is a value for style property `style_prop`.
    '''
    return style_prop in self._styles

  def get_style(self, style_prop: typing.Type[StyleProperty]):
    '''Returns the value for the style property `style_prop`, or None.'''
    return self._styles.get(style_prop)

  def set_style(self, style_prop: typing.Type[StyleProperty], value: typing.Any):
    '''Sets the value for the style property `style_prop` to `value`.'''
    if style_prop not in StyleProperties.ALL:
      raise ValueError("Invalid style property")

    if value is None:

      self._styles.pop(style_prop, None)

    else:

      if not style_prop.validate(value):
        raise ValueError(f"Invalid value {value} for style property {style_prop}")

      self._styles[style_prop] = value

  _applicableStyles: typing.Set[StyleProperty] = frozenset()

  def is_style_applicable(self, style_prop: typing.Type[StyleProperty]) -> bool:
    '''Whether the style property `style_prop` apply to the element'''
    return style_prop in self._applicableStyles

  def iter_styles(self) -> typing.Iterator[typing.Type[StyleProperty]]:
    '''Returns an iterator over all style properties that are not `None`
    '''
    return iter(self._styles)

  # layout properties

  def set_region(self, region: typing.Optional[Region]):
    '''Sets the region associated with the element, or None if the element
    is not associated with any region.'''

    if region is not None:
      if self.get_doc() is None:
        raise ValueError("Not associated with a document")
        
      if not self.get_doc().has_region(region.get_id()):
        raise ValueError("Region is unknown")

    self._region = region

  def get_region(self) -> typing.Optional[Region]:
    '''Returns the region associated with the element, or None if the element is
    not associated with any region.'''
    return self._region

  # timing properties

  def set_begin(self, time_offset: typing.Optional[Fraction]):
    '''Sets the (inclusive) begin time of the element, in seconds.'''
    self._begin = time_offset

  def get_begin(self) -> typing.Optional[Fraction]:
    '''Returns the (inclusive) begin time of the element, in seconds.'''
    return self._begin

  def set_end(self, time_offset: typing.Optional[Fraction]):
    '''Sets the (exclusive) end time of the element, in seconds.'''
    self._end = time_offset

  def get_end(self) -> typing.Optional[Fraction]:
    '''Returns the (exclusive) end time of the element, in seconds.'''
    return self._end

  # discrete animation

  def add_animation_step(self, step: DiscreteAnimationStep):
    '''Adds an animation step to the element
    '''
    if not isinstance(step, DiscreteAnimationStep):
      raise TypeError("Must be a discrete animation step")

    if not step.style_property.is_animatable:
      raise TypeError("The style property is not animatable")

    self._sets.append(step)

  def remove_animation_step(self, step: DiscreteAnimationStep):
    '''Remove `step` from the discrete animation steps associated with the element
    '''
    self._sets.remove(step)

  def iter_animation_steps(self) -> typing.Iterator[DiscreteAnimationStep]:
    '''Returns an iterator over the discrete animation steps associated with the element
    '''
    return iter(self._sets)


  # white space handling

  def set_space(self, wsh: WhiteSpaceHandling):
    '''Sets the white space handling strategy for text nodes of the element.'''
    if wsh not in WhiteSpaceHandling.__members__.values():
      raise TypeError("Must be a WhiteSpaceHandling value")

    self._space = wsh

  def get_space(self) -> WhiteSpaceHandling:
    '''Returns the white space handling strategy for text nodes of the element.'''
    return self._space

  # language

  def set_lang(self, language: str):
    '''Sets the language of the element, as an RFC 5646 language tag.'''
    self._lang = str(language)

  def get_lang(self) -> str:
    '''Returns the language of the element, as an RFC 5646 language tag.'''
    return self._lang


  
class Body(ContentElement):
  '''Body element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Display,
    StyleProperties.Opacity,
    StyleProperties.Visibility
    ])

  def push_child(self, child):
    if not isinstance(child, Div):
      raise TypeError("Children of body must be div instances")
    super().push_child(child)


  
class Div(ContentElement):
  '''Div element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Display,
    StyleProperties.Opacity,
    StyleProperties.Visibility
    ])

  def push_child(self, child):
    if not isinstance(child, (P, Div)):
      raise TypeError("Children of body must be P instances")
    super().push_child(child)


  
class P(ContentElement):
  '''P element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.FillLineGap,
    StyleProperties.FontFamily,
    StyleProperties.FontSize,
    StyleProperties.FontStyle,
    StyleProperties.FontVariant,
    StyleProperties.FontWeight,
    StyleProperties.LineHeight,
    StyleProperties.LinePadding,
    StyleProperties.MultiRowAlign,
    StyleProperties.Opacity,
    StyleProperties.RubyReserve,
    StyleProperties.Shear,
    StyleProperties.TextAlign,
    StyleProperties.UnicodeBidi,
    StyleProperties.Visibility
    ])

  def push_child(self, child):
    if not isinstance(child, (Span, Br, Ruby)):
      raise TypeError("Children of p must be span, br or ruby instances")
    super().push_child(child)


  
class Span(ContentElement):
  '''Span element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Color,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.FontFamily,
    StyleProperties.FontSize,
    StyleProperties.FontStyle,
    StyleProperties.FontVariant,
    StyleProperties.FontWeight,
    StyleProperties.Opacity,
    StyleProperties.TextCombine,
    StyleProperties.TextDecoration,
    StyleProperties.TextEmphasis,
    StyleProperties.TextOutline,
    StyleProperties.TextShadow,
    StyleProperties.UnicodeBidi,
    StyleProperties.Visibility,
    StyleProperties.WrapOption
    ])

  def push_child(self, child):
    if not isinstance(child, (Span, Br, Text)):
      raise TypeError("Children of span must be span or br instances")
    super().push_child(child)



class Br(ContentElement):
  '''Br element, as specified in TTML2'''

  def push_child(self, child):
    raise TypeError("Br elements cannot have children")

  def set_begin(self, time_offset):
    raise RuntimeError("Br elements do not have temporal properties")

  def set_end(self, time_offset):
    raise RuntimeError("Br elements do not have temporal properties")

  def set_region(self, region):
    raise RuntimeError("Br elements are not associated with a region")

  def copy_to(self, dest: Br):
    dest.set_id(self.get_id())
    dest.set_lang(self.get_lang())
    dest.set_space(self.get_space())
    
    for style_prop in self.iter_styles():
      dest.set_style(style_prop, self.get_style(style_prop))

    for anim_step in self.iter_animation_steps():
      dest.add_animation_step(anim_step)


class Ruby(ContentElement):
  '''Ruby element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.Opacity,
    StyleProperties.RubyAlign,
    StyleProperties.Visibility,
    ])

  def push_child(self, child: ContentElement):
    raise RuntimeError("Ruby children must be added using `push_children`")

  def remove_child(self, child: ContentElement):
    raise RuntimeError("Ruby children must be removed using `remove_children`")

  def push_children(self, children: typing.Iterable[ContentElement]):
    if self.has_children():
      raise RuntimeError("Remove all ruby children before adding more.")

    ts = [type(x) for x in children]

    if ts not in [[Rb, Rt], [Rb, Rp, Rt, Rp], [Rbc, Rtc], [Rbc, Rtc, Rtc]]:
      raise ValueError("Children of ruby do not conform to requirements")

    for child in children:
      super().push_child(child)

  def remove_children(self):
    '''Remove all children of the element.'''

    for child in list(self):
      super().remove_child(child)

class Rb(ContentElement):
  '''Rb element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Color,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.FontFamily,
    StyleProperties.FontSize,
    StyleProperties.FontStyle,
    StyleProperties.FontWeight,
    StyleProperties.Opacity,
    StyleProperties.TextCombine,
    StyleProperties.TextDecoration,
    StyleProperties.TextEmphasis,
    StyleProperties.TextOutline,
    StyleProperties.TextShadow,
    StyleProperties.UnicodeBidi,
    StyleProperties.Visibility,
    StyleProperties.WrapOption
    ])

  def push_child(self, child):
    if not isinstance(child, Span):
      raise TypeError("Children of rb must be span instances")
    super().push_child(child)

class Rbc(ContentElement):
  '''Rbc element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.Opacity,
    StyleProperties.Visibility,
    ])

  def push_child(self, child):
    if not isinstance(child, Rb):
      raise TypeError("Children of rc must be rb instances")
    super().push_child(child)

class Rp(ContentElement):
  '''Rp element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Color,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.FontFamily,
    StyleProperties.FontSize,
    StyleProperties.FontStyle,
    StyleProperties.FontWeight,
    StyleProperties.Opacity,
    StyleProperties.TextCombine,
    StyleProperties.TextDecoration,
    StyleProperties.TextEmphasis,
    StyleProperties.TextOutline,
    StyleProperties.TextShadow,
    StyleProperties.UnicodeBidi,
    StyleProperties.Visibility,
    StyleProperties.WrapOption
    ])

  def push_child(self, child):
    if not isinstance(child, Span):
      raise TypeError("Children of rp must be span instances")
    super().push_child(child)


class Rt(ContentElement):
  '''Rt element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Color,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.FontFamily,
    StyleProperties.FontSize,
    StyleProperties.FontStyle,
    StyleProperties.FontWeight,
    StyleProperties.Opacity,
    StyleProperties.RubyPosition,
    StyleProperties.TextCombine,
    StyleProperties.TextDecoration,
    StyleProperties.TextEmphasis,
    StyleProperties.TextOutline,
    StyleProperties.TextShadow,
    StyleProperties.UnicodeBidi,
    StyleProperties.Visibility,
    StyleProperties.WrapOption
    ])

  def push_child(self, child):
    if not isinstance(child, Span):
      raise TypeError("Children of rt must be span instances")
    super().push_child(child)


class Rtc(ContentElement):
  '''Rtc element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Direction,
    StyleProperties.Display,
    StyleProperties.Opacity,
    StyleProperties.RubyPosition,
    StyleProperties.Visibility,
    ])

  def push_child(self, child):

    if isinstance(self.first_child(), Rt):
      expect = (Rt,)
    elif isinstance(self.first_child(), Rp) and isinstance(self.last_child(), Rp):
      expect = (type(None),)
    else:
      expect = (Rt, Rp)

    if not isinstance(child, expect):
      raise ValueError("Children of rtc do not conform to requirements")

    super().push_child(child)

  def remove_child(self, child: ContentElement):
    raise RuntimeError("Rtc children must be removed using `remove_children`")

  def push_children(self, children: typing.Iterable[ContentElement]):
    cs = list(children)

    if len(cs) > 2 and isinstance(cs[0], Rp) and isinstance(cs[-1], Rp):
      cs = cs[1:-1]

    if not all(isinstance(x, Rt) for x in cs):
      raise ValueError("Children of rtc do not conform to requirements")

    for child in children:
      super().push_child(child)

  def remove_children(self):

    for child in list(self):
      super().remove_child(child)


class Text(ContentElement):
  '''Text node, as specified in TTML2'''

  def __init__(self, doc=None, text=""):
    self._text = text
    super().__init__(doc=doc)

  def copy_to(self, dest: Text):
    dest.set_text(self.get_text())

  # children

  def push_child(self, child):
    raise RuntimeError("Text nodes cannot have children")

  def set_style(self, style_prop, value):
    if value is not None:
      raise RuntimeError("Text nodes cannot have style properties")

  def set_begin(self, time_offset):
    if time_offset is not None:
      raise RuntimeError("Text nodes do not have temporal properties")

  def set_end(self, time_offset):
    if time_offset is not None:
      raise RuntimeError("Text nodes do not have temporal properties")

  def set_id(self, element_id):
    if element_id is not None:
      raise RuntimeError("Text nodes do not have an id")

  def set_region(self, region):
    if region is not None:
      raise RuntimeError("Text nodes are not associated with a region")

  def set_lang(self, language):
    if language is not None and language != "":
      raise RuntimeError("Text nodes are not associated with a language")

  def set_space(self, wsh):
    if wsh is not None and wsh is not WhiteSpaceHandling.DEFAULT:
      raise RuntimeError("Text nodes are not associated with space handling")

  # text contents

  def set_text(self, text: str):
    '''Set the text contents of the node'''
    if not isinstance(text, str):
      raise TypeError("Text must be a string")
    self._text = text

  def get_text(self) -> str:
    '''Get the text contents of the node'''
    return self._text



class Region(ContentElement):
  '''Out-of-line region element, as specified in TTML2'''

  _applicableStyles = frozenset([
    StyleProperties.BackgroundColor,
    StyleProperties.Disparity,
    StyleProperties.Display,
    StyleProperties.DisplayAlign,
    StyleProperties.Extent,
    StyleProperties.LuminanceGain,
    StyleProperties.Opacity,
    StyleProperties.Origin,
    StyleProperties.Overflow,
    StyleProperties.Padding,
    StyleProperties.Position,
    StyleProperties.ShowBackground,
    StyleProperties.Visibility,
    StyleProperties.WritingMode
    ])

  def __init__(self, region_id: str, doc: ContentDocument = None):
    super().__init__(doc)
    
    if region_id is None:
      raise ValueError("Every region must have an id")

    self._id = str(region_id)

  def copy_to(self, dest: Region):
    dest.set_lang(self.get_lang())
    dest.set_space(self.get_space())

    dest.set_begin(self.get_begin())
    dest.set_end(self.get_end())
    
    for style_prop in self.iter_styles():
      dest.set_style(style_prop, self.get_style(style_prop))

    for anim_step in self.iter_animation_steps():
      dest.add_animation_step(anim_step)

  def set_id(self, element_id):
    if element_id != self.get_id():
      raise RuntimeError("Region id is immutable")

  def push_child(self, child):
    raise RuntimeError("Region elements do not have children")

  def set_region(self, region):
    if region is not None:
      raise RuntimeError("Region elements cannot be associated with regions")
    super().set_region(region)

#
# Document
#

@dataclass(frozen=True)
class CellResolutionType:
  '''Value of the ttp:cellResolution attribute'''

  rows: int
  columns: int

  def __post_init__(self):
    if self.rows <= 0  or self.columns <= 0:
      raise ValueError("Rows and columns must be larger than 0")


@dataclass(frozen=True)
class PixelResolutionType:
  '''Extent of the root container in pixels'''

  width: int
  height: int

  def __post_init__(self):
    if self.height <= 0  or self.width <= 0:
      raise ValueError("Height and width must be larger than 0")


@dataclass(frozen=True)
class ActiveAreaType:
  '''Active area within the root container, measured as a fraction of the extent of the root container'''

  left_offset: float = 0
  top_offset: float = 0
  width: float = 1
  height: float = 1

  def __post_init__(self):
    if self.left_offset < 0  or self.left_offset > 1:
      raise ValueError("left_offset must be in the range [0, 1]")

    if self.top_offset < 0  or self.top_offset > 1:
      raise ValueError("top_offset must be in the range [0, 1]")

    if self.width < 0  or self.width > 1:
      raise ValueError("width must be in the range [0, 1]")

    if self.height < 0  or self.height > 1:
      raise ValueError("height must be in the range [0, 1]")

class Document:
  '''Base class for TTML documents, including ISDs, as specified in TTML2'''

  def __init__(self):
    self._cell_resolution = CellResolutionType(rows=15, columns=32)
    self._px_resolution = PixelResolutionType(width=1920, height=1080)
    self._active_area = None
    self._dar = None
    self._lang = ""

  
  def copy_to(self, dest: Document):
    '''Copy all information but children to a different instance.
    '''

    if dest is self:
      return

    dest.set_active_area(self.get_active_area())
    dest.set_cell_resolution(self.get_cell_resolution())
    dest.set_display_aspect_ratio(self.get_display_aspect_ratio())
    dest.set_lang(self.get_lang())
    dest.set_px_resolution(self.get_px_resolution())


  # active area

  def get_active_area(self) -> typing.Optional[ActiveAreaType]:
    '''Returns the active area of the document.
    '''
    return self._active_area

  def set_active_area(self, active_area: typing.Optional[ActiveAreaType]):
    '''Sets the active area of the document, or clears it if `None` is passed.
    '''
    if active_area is not None and not isinstance(active_area, ActiveAreaType):
      raise TypeError("Argument must be None or an instance of ActiveAreaType")

    self._active_area = active_area

  # language

  def set_lang(self, language: str):
    '''Sets the language of the document, as an RFC 5646 language tag.'''
    if not isinstance(language, str):
      raise TypeError("Argument must be a string")
    self._lang = language

  def get_lang(self) -> str:
    '''Returns the language of the document, as an RFC 5646 language tag.'''
    return self._lang

  # display aspect ratio

  def get_display_aspect_ratio(self) -> typing.Optional[Fraction]:
    '''Returns the display aspect ratio of the document, by default `None`
    '''
    return self._dar

  def set_display_aspect_ratio(self, dar: typing.Optional[Fraction]):
    '''Sets the display aspect ratio of the document to `dar`. If `dar` is `None`, the
    document will fill the root container area.
    '''
    if dar is not None and not isinstance(dar, Fraction):
      raise TypeError("Argument must be an instance of Fraction or None")

    self._dar = dar

  # cell resolution

  def get_cell_resolution(self) -> CellResolutionType:
    '''Returns the cell resolution of the document, by default 15 rows by 32 columns.'''
    return self._cell_resolution

  def set_cell_resolution(self, cell_resolution: CellResolutionType):
    '''Sets the cell resolution of the document.'''
    if not isinstance(cell_resolution, CellResolutionType):
      raise TypeError("Argument must be an instance of CellResolutionType")

    self._cell_resolution = cell_resolution

  # pixel resolution

  def get_px_resolution(self) -> PixelResolutionType:
    '''Returns the pixel resolution of the document, by default 1920 by 1080.'''
    return self._px_resolution

  def set_px_resolution(self, px_resolution: PixelResolutionType):
    '''Sets the pixel resolution of the document'''
    if not isinstance(px_resolution, PixelResolutionType):
      raise TypeError("Argument must be an instance of PixelResolutionType")

    self._px_resolution = px_resolution

class ContentDocument(Document):
  '''TTML document'''

  def __init__(self):
    self._regions = {}
    self._body = None
    self._initial_values = {}
    self._content_profiles = set()
    super().__init__()

  def copy_to(self, dest: ContentDocument):
    super().copy_to(dest)

    for style_prop, style_value in self.iter_initial_values():
      dest.put_initial_value(style_prop, style_value)

  # attributes

  def get_content_profiles(self) -> typing.Optional[typing.Set[str]]:
    '''Returns the set of content profile designators to which the document conforms. Equivalent to ttp:contentProfiles.'''
    return self._content_profiles

  def set_content_profiles(self, content_profiles: typing.Optional[typing.Set[str]]):
    '''Set the set of content profile designators to which the document conforms.'''
    if content_profiles is not None:
      if not isinstance(content_profiles, set) or not all(isinstance(e, str) and re.fullmatch(r"\S+", e) for e in content_profiles):
        raise TypeError("The content profiles must be a set of strings")

    self._content_profiles = content_profiles

  # body

  def get_body(self) -> typing.Optional[Body]:
    '''Returns the body element of the document, or None'''
    return self._body

  def set_body(self, body: typing.Optional[Body]):
    '''Sets the body element of the document, which may be None'''
    if body is not None:
      if not isinstance(body, Body):
        raise TypeError("Argument must be an instance of Body")

      if body.parent() is not None:
        raise ValueError("Body must be a root element")

      if body.get_doc() is not self:
        raise ValueError("Body does not belongs to this document")

    self._body = body

  # regions

  def has_region(self, region_id: str) -> bool:
    '''Returns whether the document has a region with an id of `region_id`.'''
    return region_id in self._regions

  def put_region(self, region: Region):
    '''Adds a region to the document, replacing any existing region with the same `id`.'''
    if not isinstance(region, Region):
      raise TypeError("Argument must be an instance of Region")

    if region.get_doc() != self:
      raise ValueError("Region does not belongs to this document")

    self._regions[region.get_id()] = region

  def remove_region(self, region_id: str):
    '''Removes the region with `id == region_id` from the document and all content elements.'''
    region = self.get_region(region_id)

    if region is None:
      return

    # removes the region from all content elements

    body = self.get_body()

    if body is not None: 
      map(
        lambda e: e.get_region() and e.get_region().get_id() == region_id and e.set_region(None),
        body.dfs_iterator()
      )

    del self._regions[region_id]

  def get_region(self, region_id) -> typing.Optional[Region]:
    '''Returns the region with `id == region_id` or None, if none exists.'''
    return self._regions.get(region_id)

  def iter_regions(self) -> typing.Iterator[Region]:
    '''Returns an iterator over regions.'''
    return self._regions.values()

  # initial value

  def get_initial_value(self, style_prop: StyleProperty) -> typing.Any:
    '''Returns the initial value for the style property `style_prop`, or None otherwise.'''
    return self._initial_values.get(style_prop)

  def put_initial_value(self, style_prop: StyleProperty, initial_value: typing.Any):
    '''Adds an initial value for the style property `style_prop`,
    replacing any existing one for the same property. If `initial_value` is `None`,
    the initial value is removed.'''
    if style_prop not in StyleProperties.ALL:
      raise ValueError("Invalid style property")

    if initial_value is None:

      self.remove_initial_value(style_prop)

    else:

      if not style_prop.validate(initial_value):

        raise ValueError("Invalid value")

      self._initial_values[style_prop] = initial_value

  def remove_initial_value(self, style_prop: StyleProperty):
    '''Removes any initial value for the style property `style_prop`.'''
    self._initial_values.pop(style_prop, None)

  def has_initial_value(self, style_prop: StyleProperty) -> typing.Any:
    '''Returns whether the document has an initial value for the style property `style_prop`.'''
    return style_prop in self._initial_values

  def iter_initial_values(self) -> typing.Iterator[typing.Tuple[StyleProperty, typing.Any]]:
    '''Returns an iterator over (style property, initial value) pairs.'''
    return self._initial_values.items()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Style properties'''

from __future__ import annotations
import typing
from dataclasses import dataclass
import numbers
from typing import Optional, SupportsFloat as Numeric
from enum import Enum

#
# Types
#

class AnnotationPositionType(Enum):
  '''<\\annotation-position\\> value
  '''
  before = "before"
  after = "after"
  outside = "outside" 


@dataclass(frozen=True)
class LengthType:
  '''Length type as defined in TTML
  '''

  class Units(Enum):
    '''Units of length
    '''
    em = "em"
    pct = "%"
    rh = "rh"
    rw = "rw"
    c = "c"
    px = "px"

  value: Numeric = 0
  units: Units = Units.pct

  def __post_init__(self):
    if not isinstance(self.value, numbers.Number):
      raise ValueError("The length value must be a number")

    if not isinstance(self.units, LengthType.Units):
      raise ValueError("Invalid units")

  def is_non_negative(self):
    '''True if the length value is positive
    '''
    return self.value >= 0

@dataclass(frozen=True)
class ColorType:
  '''<color> type as defined in TTML
  '''

  class Colorimetry(Enum):
    '''Supported colorimetry systems
    '''
    RGBA8 = "RGBA8"

  components: tuple = (255, 255, 255, 255)
  ident: Colorimetry = Colorimetry.RGBA8


class NamedColors(Enum):
  '''TTML \\<named-color\\> 
  '''
  transparent = ColorType((0, 0, 0, 0))
  black = ColorType((0, 0, 0, 255))
  silver = ColorType((192, 192, 192, 255))
  gray = ColorType((128, 128, 128, 255))
  white = ColorType((255, 255, 255, 255))
  maroon = ColorType((128, 0, 0, 255))
  red = ColorType((255, 0, 0, 255))
  purple = ColorType((128, 0, 128, 255))
  fuchsia = ColorType((255, 0, 255, 255))
  magenta = ColorType((255, 0, 255, 255))
  green = ColorType((0, 128, 0, 255))
  lime = ColorType((0, 255, 0, 255))
  olive = ColorType((128, 128, 0, 255))
  yellow = ColorType((255, 255, 0, 255))
  navy = ColorType((0, 0, 128, 255))
  blue = ColorType((0, 0, 255, 255))
  teal = ColorType((0, 128, 128, 255))
  aqua = ColorType((0, 255, 255, 255))
  cyan = ColorType((0, 255, 255, 255))


class DirectionType(Enum):
  '''tts:direction values
  '''
  ltr = "ltr"
  rtl = "rtl" 


class DisplayType(Enum):
  '''tts:display value
  '''
  auto = "auto"
  none = "none" 


class DisplayAlignType(Enum):
  '''tts:displayAlign value
  '''
  before = "before"
  center = "center"
  after = "after" 

@dataclass(frozen=True)
class ExtentType:
  '''tts:extent value
  '''

  height: LengthType = LengthType()
  width: LengthType = LengthType()

  def __post_init__(self):
    if not isinstance(self.height, LengthType) or not isinstance(self.width, LengthType):
      raise ValueError("Dimensions must be a length")

class GenericFontFamilyType(Enum):
  '''\\<generic-family-name\\>
  '''
  default = "default"
  monospace = "monospace"
  sansSerif = "sansSerif"
  serif = "serif"
  monospaceSansSerif = "monospaceSansSerif"
  monospaceSerif = "monospaceSerif"
  proportionalSansSerif = "proportionalSansSerif"
  proportionalSerif = "proportionalSerif"

class FontStyleType(Enum):
  '''tts:fontStyle value
  '''
  normal = "normal"
  italic = "italic"
  oblique = "oblique" 


class FontVariantType(Enum):
  '''tts:fontVariant value
  '''
  normal = "normal"
  superscript = "super"
  subscript = "sub" 


class FontWeightType(Enum):
  '''tts:fontWeight value
  '''
  normal = "normal"
  bold = "bold" 


class MultiRowAlignType(Enum):
  '''ebutts:multiRowAlign value
  '''
  start = "start"
  center = "center"
  end = "end"
  auto = "auto"

class OverflowType(Enum):
  '''tts:overflow value
  '''
  visible = "visible"
  hidden = "hidden" 

@dataclass(frozen=True)
class PaddingType:
  '''tts:padding value
  '''

  before: LengthType = LengthType()
  end: LengthType = LengthType()
  after: LengthType = LengthType()
  start: LengthType = LengthType()

  def __post_init__(self):
    if not isinstance(self.before, LengthType) or not isinstance(self.end, LengthType) \
      or not isinstance(self.after, LengthType) or not isinstance(self.start, LengthType):
      raise ValueError("Dimensions must be a length")

class RubyAlignType(Enum):
  '''tts:rubyAlign value
  '''
  center = "center"
  spaceAround = "spaceAround" 


@dataclass(frozen=True)
class RubyReserveType:
  '''TTML \\<ruby-reserve\\>
  '''

  class Position(Enum):
    '''TTML \\<ruby-reserve\\> position'''
    both = "both"
    before = "before"
    after = "after"
    outside = "outside" 

  position: Position = Position.outside
  length: Optional[LengthType] = None

  def __post_init__(self):
    if self.length is not None and not isinstance(self.length, LengthType):
      raise ValueError("Length must be a length")

class ShowBackgroundType(Enum):
  '''tts:showBackground values
  '''
  always = "always"
  whenActive = "whenActive"

class SpecialValues(Enum):
  '''Special style property values
  '''
  normal = "normal"
  none = "none"

class TextAlignType(Enum):
  '''tts:textAlign value
  '''
  center = "center"
  start = "start"
  end = "end"


class TextCombineType(Enum):
  '''TTML \\<text-combine\\>
  '''
  none = "none"
  all = "all"


@dataclass(frozen=True)
class TextDecorationType:
  '''TTML \\<text-decoration\\>
  '''

  underline: typing.Optional[bool] = None
  line_through: typing.Optional[bool] = None
  overline: typing.Optional[bool] = None


@dataclass(frozen=True)
class TextEmphasisType:
  '''TTML \\<text-emphasis\\>
  '''

  class Style(Enum):
    '''Combines \\<emphasis-style\\> and \\<emphasis-symbol\\>
    '''
    auto = "auto"
    filled_circle = "filled circle"
    filled_dot = "filled dot"
    filled_sesame = "filled sesame"
    open_circle = "open circle"
    open_dot = "open dot"
    open_sesame = "open sesame"

  class Position(Enum):
    ''' \\<emphasis-position\\>
    '''
    outside = "outside"
    before = "before"
    after = "after"

  style: Style = Style.auto
  color: Optional[ColorType] = None
  position: Position = Position.outside

  def __post_init__(self):
    if not isinstance(self.style, TextEmphasisType.Style):
      raise ValueError("Style must be a text emphasis style enumeration value")

    if not (self.color is None or isinstance(self.color, ColorType)):
      raise ValueError("The color must be None or a valid color")

    if not isinstance(self.position, TextEmphasisType.Position):
      raise ValueError("Position must be a text emphasis position enumeration value")


@dataclass(frozen=True)
class TextOutlineType:
  '''TTML \\<text-outline\\>
  '''

  thickness: LengthType
  color: Optional[ColorType] = None

  def __post_init__(self):
    if self.thickness is None or not isinstance(self.thickness, LengthType):
      raise ValueError("The thickness value must be a length")

    if not (self.color is None or isinstance(self.color, ColorType)):
      raise ValueError("The color must be None or a valid color")

@dataclass(frozen=True)
class TextShadowType:
  '''TTML \\<text-shadow\\>
  '''

  @dataclass(frozen=True)
  class Shadow:
    '''Represents one shadow value
    '''
    x_offset: LengthType
    y_offset: LengthType
    blur_radius: typing.Optional[LengthType] = None
    color: typing.Optional[ColorType] = None

    def __post_init__(self):
      if self.blur_radius is not None and not isinstance(self.blur_radius, LengthType):
        raise ValueError("The blur_radius value must be a length")

  shadows: typing.Tuple[TextShadowType.Shadow]

class UnicodeBidiType(Enum):
  '''tts:unicodeBidi values
  '''
  normal = "normal"
  embed = "embed"
  bidiOverride = "bidiOverride"


class VisibilityType(Enum):
  '''tts:visibility value
  '''
  visible = "visible"
  hidden = "hidden" 


class WrapOptionType(Enum):
  '''tts:wrapOptionType value
  '''
  wrap = "wrap"
  noWrap = "noWrap" 


class WritingModeType(Enum):
  '''tts:writingModeType value
  '''
  lrtb = "lrtb"
  rltb = "rltb"
  tbrl = "tbrl" 
  tblr = "tblr" 


@dataclass(frozen=True)
class CoordinateType:
  '''Coordinates (`x`, `y`) in the root container region, as measure from the left and top edges, respectively.
  '''
  x: LengthType
  y: LengthType

  def __post_init__(self):
    if not isinstance(self.x, LengthType) or not isinstance(self.y, LengthType):
      raise ValueError("Coordinates must be a length")

@dataclass(frozen=True)
class PositionType:
  '''Offsets (`h_offset`, `v_offset`) in the root container region, as measure from the `h_edge` and `v_edge` edges, respectively.
  '''
  class HEdge(Enum):
    left = "left"
    right = "right"

  class VEdge(Enum):
    top = "top"
    bottom = "bottom"

  h_offset: LengthType
  v_offset: LengthType
  h_edge: HEdge = HEdge.left
  v_edge: VEdge = VEdge.top

  def __post_init__(self):
    if not isinstance(self.h_offset, LengthType) or not isinstance(self.v_offset, LengthType):
      raise ValueError("Offset must be a length")

    if LengthType.Units.c in (self.h_offset.units, self.v_offset.units):
      raise ValueError("Offset must not be in 'c' units")

#
# Style properties
#

class StyleProperty:
  '''Abstract base class for all style properties'''

  @classmethod
  def is_inherited(cls) -> bool:
    '''True if a child element inherits the value of the property from its parent
    '''
    raise NotImplementedError

  @classmethod
  def is_animatable(cls) -> bool:
    '''True if the property can be use in a `DiscreteAnimationStep` instance
    '''
    raise NotImplementedError

  @staticmethod
  def validate(value: typing.Any) -> bool:
    '''Returns whether the value is valid for the style property.'''
    raise NotImplementedError

  @staticmethod
  def make_initial_value() -> typing.Any:
    '''Creates an instance of the initial value of the style property.'''
    raise NotImplementedError

class StyleProperties:
  '''Container for all style properties
  
  Class variables:

  * `ALL`: set of all style properties
  '''

  class BackgroundColor(StyleProperty):
    '''Corresponds to tts:backgroundColor.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return NamedColors.transparent.value

    @staticmethod
    def validate(value):
      return isinstance(value, ColorType)


  class Color(StyleProperty):
    '''Corresponds to tts:color.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return NamedColors.white.value

    @staticmethod
    def validate(value):
      return isinstance(value, ColorType) 


  class Direction(StyleProperty):
    '''Corresponds to tts:direction.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return DirectionType.ltr

    @staticmethod
    def validate(value):
      return isinstance(value, DirectionType) 

  class Disparity(StyleProperty):
    '''Corresponds to tts:disparity.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return LengthType()

    @staticmethod
    def validate(value):
      return isinstance(value, LengthType) and value.units != LengthType.Units.c

  class Display(StyleProperty):
    '''Corresponds to tts:display.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return DisplayType.auto

    @staticmethod
    def validate(value):
      return isinstance(value, DisplayType) 


  class DisplayAlign(StyleProperty):
    '''Corresponds to tts:displayAlign.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return DisplayAlignType.before

    @staticmethod
    def validate(value):
      return isinstance(value, DisplayAlignType) 


  class Extent(StyleProperty):
    '''Corresponds to tts:extent.
    '''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return ExtentType(
        height=LengthType(value=100, units=LengthType.Units.pct),
        width=LengthType(value=100, units=LengthType.Units.pct)
      )

    @staticmethod
    def validate(value: ExtentType):
      return isinstance(value, ExtentType) \
        and value.width.units in (LengthType.Units.pct, LengthType.Units.px, LengthType.Units.rw)  \
        and value.height.units in (LengthType.Units.pct, LengthType.Units.px, LengthType.Units.rh) \
        and value.width.is_non_negative() and value.height.is_non_negative()


  class FillLineGap(StyleProperty):
    '''Corresponds to itts:fillLineGap.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return False

    @staticmethod
    def validate(value: bool):
      return isinstance(value, bool)


  class FontFamily(StyleProperty):
    '''Corresponds to tts:fontFamily.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return (GenericFontFamilyType.default,)

    @staticmethod
    def validate(value: typing.Tuple[typing.Union[str, GenericFontFamilyType]]):
      return isinstance(value, tuple) and all(lambda i: isinstance(i, (str, GenericFontFamilyType)) for i in value)


  class FontSize(StyleProperty):
    '''Corresponds to tts:fontSize.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return LengthType(1, LengthType.Units.c)

    @staticmethod
    def validate(value):
      return isinstance(value, LengthType) and value.is_non_negative()


  class FontStyle(StyleProperty):
    '''Corresponds to tts:fontStyle.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return FontStyleType.normal

    @staticmethod
    def validate(value):
      return isinstance(value, FontStyleType)


  class FontVariant(StyleProperty):
    '''Corresponds to tts:fontVariant.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return FontVariantType.normal

    @staticmethod
    def validate(value):
      return isinstance(value, FontVariantType)


  class FontWeight(StyleProperty):
    '''Corresponds to tts:fontWeight.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return FontWeightType.normal

    @staticmethod
    def validate(value):
      return isinstance(value, FontWeightType)


  class LineHeight(StyleProperty):
    '''Corresponds to tts:lineHeight.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return SpecialValues.normal

    @staticmethod
    def validate(value):
      return value == SpecialValues.normal or \
        (isinstance(value, LengthType) and value.units != LengthType.Units.c and value.is_non_negative())


  class LinePadding(StyleProperty):
    '''Corresponds to ebutts:linePadding.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return LengthType(value=0, units=LengthType.Units.c)

    @staticmethod
    def validate(value: LengthType):
      return isinstance(value, LengthType) and \
        value.units in (LengthType.Units.c, LengthType.Units.rh, LengthType.Units.rw) and \
        value.is_non_negative()

  class LuminanceGain(StyleProperty):
    '''Corresponds to tts:luminanceGain.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return 1.0

    @staticmethod
    def validate(value: Numeric):
      return isinstance(value, numbers.Number) and value >= 0

  class MultiRowAlign(StyleProperty):
    '''Corresponds to ebutts:multiRowAlign.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return MultiRowAlignType.auto

    @staticmethod
    def validate(value):
      return isinstance(value, MultiRowAlignType)


  class Opacity(StyleProperty):
    '''Corresponds to tts:opacity.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return 1.0

    @staticmethod
    def validate(value: Numeric):
      return isinstance(value, numbers.Number)


  class Origin(StyleProperty):
    '''Corresponds to tts:origin.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return CoordinateType(
        LengthType(0, LengthType.Units.pct),
        LengthType(0, LengthType.Units.pct)
      )

    @staticmethod
    def validate(value: CoordinateType):
      return isinstance(value, CoordinateType) \
        and value.x.units in (LengthType.Units.pct, LengthType.Units.px, LengthType.Units.rh, LengthType.Units.rw)  \
        and value.y.units in (LengthType.Units.pct, LengthType.Units.px, LengthType.Units.rh, LengthType.Units.rw) \
        and value.x.is_non_negative() and value.y.is_non_negative()


  class Overflow(StyleProperty):
    '''Corresponds to tts:overflow.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return OverflowType.hidden

    @staticmethod
    def validate(value):
      return isinstance(value, OverflowType)


  class Padding(StyleProperty):
    '''Corresponds to tts:padding.'''
    
    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return PaddingType()

    @staticmethod
    def validate(value):
      if not isinstance(value, PaddingType):
        return False

      return LengthType.Units.c not in (value.before.units, value.end.units, value.after.units, value.start.units) and \
        value.before.is_non_negative() and value.end.is_non_negative() and value.after.is_non_negative() and value.start.is_non_negative()

  class Position(StyleProperty):
    '''Corresponds to tts:position.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return PositionType(
        LengthType(0, LengthType.Units.pct),
        LengthType(0, LengthType.Units.pct)
      )

    @staticmethod
    def validate(value: PositionType):
      return isinstance(value, PositionType) \
        and value.h_offset.units in (LengthType.Units.pct, LengthType.Units.px, LengthType.Units.rh, LengthType.Units.rw)  \
        and value.v_offset.units in (LengthType.Units.pct, LengthType.Units.px, LengthType.Units.rh, LengthType.Units.rw) \
        and value.h_offset.is_non_negative() and value.v_offset.is_non_negative()

  class RubyAlign(StyleProperty):
    '''Corresponds to tts:rubyAlign.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return RubyAlignType.center

    @staticmethod
    def validate(value):
      return isinstance(value, RubyAlignType)


  class RubyPosition(StyleProperty):
    '''Corresponds to tts:rubyPosition.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return AnnotationPositionType.outside

    @staticmethod
    def validate(value):
      return isinstance(value, AnnotationPositionType)


  class RubyReserve(StyleProperty):
    '''Corresponds to tts:rubyReserve.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return SpecialValues.none

    @staticmethod
    def validate(value):
      if value == SpecialValues.none:
        return True

      if not isinstance(value, RubyReserveType):
        return False

      return value.length is None or (value.length.units != LengthType.Units.c and value.length.is_non_negative())

  class Shear(StyleProperty):
    '''Corresponds to tts:shear.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return 0.0

    @staticmethod
    def validate(value: Numeric):
      return isinstance(value, numbers.Number) and value >= -100.0 and value <= 100.0

  class ShowBackground(StyleProperty):
    '''Corresponds to tts:showBackground.'''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return ShowBackgroundType.always

    @staticmethod
    def validate(value):
      return isinstance(value, ShowBackgroundType)


  class TextAlign(StyleProperty):
    '''Corresponds to tts:textAlign.'''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return TextAlignType.start

    @staticmethod
    def validate(value):
      return isinstance(value, TextAlignType)


  class TextCombine(StyleProperty):
    '''Corresponds to tts:textCombine
    '''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return TextCombineType.none

    @staticmethod
    def validate(value):
      return isinstance(value, TextCombineType)


  class TextDecoration(StyleProperty):
    '''Corresponds to tts:textDecoration
    '''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return TextDecorationType(
        underline=False,
        overline=False,
        line_through=False
      )

    @staticmethod
    def validate(value):
      return isinstance(value, TextDecorationType)


  class TextEmphasis(StyleProperty):
    '''Corresponds to tts:textEmphasis
    '''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return SpecialValues.none

    @staticmethod
    def validate(value):
      return value is SpecialValues.none or isinstance(value, TextEmphasisType)


  class TextOutline(StyleProperty):
    '''Corresponds to tts:textOutline
    '''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return SpecialValues.none

    @staticmethod
    def validate(value):
      if value == SpecialValues.none:
        return True

      if not isinstance(value, TextOutlineType):
        return False

      return value.thickness.units != LengthType.Units.c and value.thickness.is_non_negative()


  class TextShadow(StyleProperty):
    '''Corresponds to tts:textShadow
    '''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return SpecialValues.none

    @staticmethod
    def validate(value):
      if value == SpecialValues.none:
        return True

      if not isinstance(value, TextShadowType):
        return False

      return len(value.shadows) <= 4 and all(s.blur_radius is None or s.blur_radius.units != LengthType.Units.c for s in value.shadows)


  class UnicodeBidi(StyleProperty):
    '''Corresponds to tts:unicodeBidi
    '''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return UnicodeBidiType.normal

    @staticmethod
    def validate(value):
      return isinstance(value, UnicodeBidiType)


  class Visibility(StyleProperty):
    '''Corresponds to tts:visibility
    '''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return VisibilityType.visible

    @staticmethod
    def validate(value):
      return isinstance(value, VisibilityType)


  class WrapOption(StyleProperty):
    '''Corresponds to tts:wrapOption
    '''

    is_inherited = True
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return WrapOptionType.wrap

    @staticmethod
    def validate(value):
      return isinstance(value, WrapOptionType)


  class WritingMode(StyleProperty):
    '''Corresponds to tts:writingMode
    '''

    is_inherited = False
    is_animatable = True

    @staticmethod
    def make_initial_value():
      return WritingModeType.lrtb

    @staticmethod
    def validate(value):
      return isinstance(value, WritingModeType)

  ALL = {v: StyleProperty for n, v in list(locals().items()) if callable(v)}



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the data model'''

# pylint: disable=R0201,C0115,C0116

import glob
import typing
import unittest
import os
import logging
import xml.etree.ElementTree as et
from fractions import Fraction
import ttconv.imsc.reader as imsc_reader
import ttconv.model as model
import ttconv.style_properties as styles
from ttconv.isd import ISD

def _print_isd_node(element, level):
  if isinstance(element, model.Text):
    print('"', element.get_text(), '"')
  else:
    print(" " * 2 * level, element.__class__.__name__)

  for child in element:
    _print_isd_node(child, level + 1)

class ContentDocument0Test(unittest.TestCase):

  '''
  <region xml:id="r1"/>
  <region xml:id="r2" begin="2s" end="9s">
    <set tts:color="red"/>
  </region>

  <body begin="1s" end="10s">
    <div begin="3s" region="r1">
      <set begin="1s" tts:color="green"/>
    </div>
    <div end="12s" region="r2">
      <p>
        <span>
        hello
        </span>
      </p>
    </div>
  </body>
  '''

  def setUp(self):
    self.doc = model.ContentDocument()

    a1 = model.DiscreteAnimationStep(
      style_property=styles.StyleProperties.Color,
      begin=None,
      end=None,
      value=styles.NamedColors.red.value
    )

    a2 = model.DiscreteAnimationStep(
      style_property=styles.StyleProperties.Color,
      begin=Fraction(1),
      end=None,
      value=styles.NamedColors.green.value
    )

    a3 = model.DiscreteAnimationStep(
      style_property=styles.StyleProperties.Color,
      begin=Fraction(2),
      end=None,
      value=styles.NamedColors.blue.value
    )

    r1 = model.Region("r1", self.doc)
    self.doc.put_region(r1)

    # r2: sig times = {2, 9}

    r2 = model.Region("r2", self.doc)
    r2.set_begin(Fraction(2))
    r2.set_end(Fraction(9))
    r2.add_animation_step(a1)
    self.doc.put_region(r2)

    # b: sig times = {1, 10}

    b = model.Body(self.doc)
    b.set_begin(Fraction(1))
    b.set_end(Fraction(10))
    self.doc.set_body(b)

    # div1: offset = 1, sig times = {2, 4}

    div1 = model.Div(self.doc)
    div1.add_animation_step(a2)
    div1.set_begin(Fraction(3))
    div1.set_region(r1)
    b.push_child(div1)

    # div2: offset = 1, sig times = {10}

    div2 = model.Div(self.doc)
    div2.set_end(Fraction(12))
    div2.set_region(r2)
    b.push_child(div2)

    # p1: offset = 1, sig times = {}

    p1 = model.P(self.doc)
    div2.push_child(p1)

    # span1: offset = 1, sig times = {3}
    span1 = model.Span(self.doc)
    span1.add_animation_step(a3)
    p1.push_child(span1)

    t1 = model.Text(self.doc, "hello")
    span1.push_child(t1)

  def test_iter(self):
    isd = ISD.from_model(self.doc, 2)
    r_ids = sorted(r.get_id() for r in isd)
    self.assertSequenceEqual(r_ids, sorted(("r1", "r2")))

  def test_significant_times(self):
    self.assertSequenceEqual(ISD.significant_times(self.doc), sorted((0, 2, 3, 9, 1, 10, 4)))

  def test_isd_0(self):
    isd = ISD.from_model(self.doc, 0)

    regions = list(isd.iter_regions())

    self.assertEqual(len(regions), 1)

    r1 = regions[0]

    self.assertEqual(r1.get_id(), "r1")

    self.assertEqual(len(r1), 0)

  def test_isd_2(self):
    isd = ISD.from_model(self.doc, 2)

    regions = list(isd.iter_regions())

    self.assertEqual(len(regions), 2)

    r1 = regions[0]

    self.assertEqual(r1.get_id(), "r1")

    self.assertEqual(len(r1), 0)

    r2 = regions[1]

    self.assertEqual(r2.get_id(), "r2")

    r2_children = list(r2)

    self.assertEqual(len(r2_children), 1)

    body = r2_children[0]

    self.assertIsInstance(body, model.Body)

    self.assertEqual(len(body), 1)

    div = list(body)[0]

    self.assertIsInstance(div, model.Div)

    self.assertEqual(len(div), 1)

    p = list(div)[0]

    self.assertIsInstance(p, model.P)

    self.assertEqual(len(p), 1)

    span = list(p)[0]

    self.assertIsInstance(span, model.Span)

    self.assertEqual(len(span), 1)

    text = list(span)[0]

    self.assertIsInstance(text, model.Text)

    self.assertEqual(text.get_text(), "hello")

  def test_isd_10(self):
    isd = ISD.from_model(self.doc, 0)

    regions = list(isd.iter_regions())

    self.assertEqual(len(regions), 1)

    r1 = regions[0]

    self.assertEqual(r1.get_id(), "r1")

    self.assertEqual(len(r1), 0)


class IMSCTestSuiteTest(unittest.TestCase):

  def test_display_none_handling(self):
    xml_doc = et.parse("src/test/resources/ttml/imsc-tests/imsc1/ttml/timing/MediaParTiming002.ttml")
    doc = imsc_reader.to_model(xml_doc)
    isd = ISD.from_model(doc, 0)

    regions = list(isd.iter_regions())

    # single default region

    self.assertEqual(len(regions), 1)

    # no content

    self.assertEqual(len(regions[0]), 0)



  def test_imsc_1_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name), self.assertLogs() as logs:
            logging.getLogger().info("*****dummy*****") # dummy log
            tree = et.parse(os.path.join(root, filename))
            m = imsc_reader.to_model(tree)
            self.assertIsNotNone(m)
            sig_times = ISD.significant_times(m)
            for t in sig_times:
              isd = ISD.from_model(m, t)
              self.assertIsNotNone(isd)
            if len(logs.output) > 1:
              self.fail(logs.output)

  def test_imsc_1_1_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1_1/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name), self.assertLogs() as logs:
            logging.getLogger().info("*****dummy*****") # dummy log
            tree = et.parse(os.path.join(root, filename))
            m = imsc_reader.to_model(tree)
            self.assertIsNotNone(m)
            sig_times = ISD.significant_times(m)
            for t in sig_times:
              isd = ISD.from_model(m, t)
              self.assertIsNotNone(isd)
            if len(logs.output) > 1:
              self.fail(logs.output)
  
  def test_imsc_1_3_test_suite(self):
    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1_3/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name), self.assertLogs() as logs:
            logging.getLogger().info("*****dummy*****") # dummy log
            tree = et.parse(os.path.join(root, filename))
            m = imsc_reader.to_model(tree)
            self.assertIsNotNone(m)
            sig_times = ISD.significant_times(m)
            for t in sig_times:
              isd = ISD.from_model(m, t)
              self.assertIsNotNone(isd)
            if len(logs.output) > 1:
              self.fail(logs.output)

class ComputeStyleTest(unittest.TestCase):

  def test_compute_extent_pct(self):
    doc = model.ContentDocument()

    r1 = model.Region("r1", doc)
    r1.set_style(styles.StyleProperties.ShowBackground, styles.ShowBackgroundType.always)
    r1.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        width=styles.LengthType(50, styles.LengthType.Units.pct),
        height=styles.LengthType(25, styles.LengthType.Units.pct)
      )
    )
    doc.put_region(r1)

    isd = ISD.from_model(doc, 0)

    region = list(isd.iter_regions())[0]

    extent: styles.ExtentType = region.get_style(styles.StyleProperties.Extent)

    self.assertEqual(extent.height.value, 25)
    self.assertEqual(extent.height.units, styles.LengthType.Units.rh)

    self.assertEqual(extent.width.value, 50)
    self.assertEqual(extent.width.units, styles.LengthType.Units.rw)

  def test_compute_extent_px(self):
    doc = model.ContentDocument()

    r1 = model.Region("r1", doc)
    r1.set_style(styles.StyleProperties.ShowBackground, styles.ShowBackgroundType.always)
    r1.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        width=styles.LengthType(50, styles.LengthType.Units.px),
        height=styles.LengthType(25, styles.LengthType.Units.px)
      )
    )
    doc.put_region(r1)

    isd = ISD.from_model(doc, 0)

    region = list(isd.iter_regions())[0]

    extent: styles.ExtentType = region.get_style(styles.StyleProperties.Extent)

    self.assertAlmostEqual(extent.width.value, 100*50/doc.get_px_resolution().width)
    self.assertEqual(extent.width.units, styles.LengthType.Units.rw)

    self.assertAlmostEqual(extent.height.value, 100*25/doc.get_px_resolution().height)
    self.assertEqual(extent.height.units, styles.LengthType.Units.rh)

  def test_compute_extent_c(self):
    doc = model.ContentDocument()

    r1 = model.Region("r1", doc)
    r1.set_style(styles.StyleProperties.ShowBackground, styles.ShowBackgroundType.whenActive)
    doc.put_region(r1)

    b = model.Body(doc)
    b.set_region(r1)
    doc.set_body(b)

    div1 = model.Div(doc)
    b.push_child(div1)

    p1 = model.P(doc)
    p1.set_begin(0)
    p1.set_end(3)
    p1.set_style(
      styles.StyleProperties.LinePadding,
      styles.LengthType(1, styles.LengthType.Units.c)
    )

    div1.push_child(p1)

    span1 = model.Span(doc)
    span1.push_child(model.Text(doc, "hello"))
    p1.push_child(span1)

    isd = ISD.from_model(doc, 0)

    r1 = list(isd.iter_regions())[0]
    p1 = r1.first_child().first_child().first_child()

    l: styles.LengthType = p1.get_style(styles.StyleProperties.LinePadding)

    self.assertAlmostEqual(l.value, 100/doc.get_cell_resolution().rows)
    self.assertEqual(l.units, styles.LengthType.Units.rh)

  def test_compute_extent_em(self):
    doc = model.ContentDocument()

    r1 = model.Region("r1", doc)
    r1.set_style(styles.StyleProperties.ShowBackground, styles.ShowBackgroundType.always)
    with self.assertRaises(ValueError) as _context:
      r1.set_style(
        styles.StyleProperties.Extent,
        styles.ExtentType(
          width=styles.LengthType(20, styles.LengthType.Units.em),
          height=styles.LengthType(3, styles.LengthType.Units.em)
        )
      )

  def test_compute_padding(self):
    doc = model.ContentDocument()

    r1 = model.Region("r1", doc)
    r1.set_style(styles.StyleProperties.ShowBackground, styles.ShowBackgroundType.always)
    r1.set_style(
      styles.StyleProperties.Extent,
      styles.ExtentType(
        width=styles.LengthType(50, styles.LengthType.Units.pct),
        height=styles.LengthType(25, styles.LengthType.Units.pct)
      )
    )
    r1.set_style(
      styles.StyleProperties.Padding,
      styles.PaddingType(
        before=styles.LengthType(5, styles.LengthType.Units.pct),
        after=styles.LengthType(10, styles.LengthType.Units.pct),
        start=styles.LengthType(15, styles.LengthType.Units.pct),
        end=styles.LengthType(20, styles.LengthType.Units.pct)
      )
    )
    doc.put_region(r1)

    isd = ISD.from_model(doc, 0)

    region = list(isd.iter_regions())[0]

    padding: styles.PaddingType = region.get_style(styles.StyleProperties.Padding)

    self.assertAlmostEqual(padding.before.value, 25 * 0.05)
    self.assertAlmostEqual(padding.after.value, 25 * 0.10)
    self.assertAlmostEqual(padding.start.value, 50 * 0.15)
    self.assertAlmostEqual(padding.end.value, 50 * 0.2)

  def test_compute_style_property(self):
    doc = model.ContentDocument()

    r1 = model.Region("r1", doc)
    r1.set_style(styles.StyleProperties.FontSize, styles.LengthType(value=50, units=styles.LengthType.Units.pct))
    doc.put_region(r1)

    b = model.Body(doc)
    b.set_style(styles.StyleProperties.FontSize, styles.LengthType(value=50, units=styles.LengthType.Units.pct))
    b.set_region(r1)
    doc.set_body(b)

    div1 = model.Div(doc)
    b.push_child(div1)

    p1 = model.P(doc)
    div1.push_child(p1)

    span1 = model.Span(doc)
    p1.push_child(span1)

    t1 = model.Text(doc, "hello")
    span1.push_child(t1)

    isd = ISD.from_model(doc, 0)

    region = list(isd.iter_regions())[0]

    span = region[0][0][0][0]

    fs: styles.LengthType = span.get_style(styles.StyleProperties.FontSize)

    self.assertAlmostEqual(fs.value, 25 / doc.get_cell_resolution().rows)

    self.assertEqual(fs.units, styles.LengthType.Units.rh)

class InheritanceStyleTest(unittest.TestCase):

  def test_text_decoration_inheritance(self):
    doc = model.ContentDocument()

    r1 = model.Region("r1", doc)
    r1.set_style(
      styles.StyleProperties.TextDecoration,
      styles.TextDecorationType(
        line_through=False,
        underline=True,
        overline=True
      )
    )
    doc.put_region(r1)

    b = model.Body(doc)
    b.set_style(
      styles.StyleProperties.TextDecoration,
      styles.TextDecorationType(
        overline=False
      )
    )
    b.set_region(r1)
    doc.set_body(b)

    div1 = model.Div(doc)
    b.push_child(div1)

    p1 = model.P(doc)
    div1.push_child(p1)

    span1 = model.Span(doc)
    p1.push_child(span1)

    t1 = model.Text(doc, "hello")
    span1.push_child(t1)

    isd = ISD.from_model(doc, 0)

    region = list(isd.iter_regions())[0]

    span = region[0][0][0][0]

    self.assertEqual(
      span.get_style(styles.StyleProperties.TextDecoration),
      styles.TextDecorationType(
        line_through=False,
        underline=True,
        overline=False
      )
    )

  def test_textEmphasis_auto(self):
    """https://github.com/sandflow/ttconv/issues/400"""
    xml_str = """<?xml version="1.0" encoding="utf-8"?>
<tt xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling" xml:lang="en">
<head>
<styling>
<initial tts:writingMode="lrtb"/>
<style xml:id="s1" tts:textEmphasis="auto outside"/>
</styling>
</head>
<body>
<div>
<p begin="0s" end="2s"><span style="s1">hello</span></p>
</div>
</body>
</tt>"""

    tree = et.ElementTree(et.fromstring(xml_str))
    doc = imsc_reader.to_model(tree)
    isd = ISD.from_model(doc, 0)

    regions = list(isd.iter_regions())

    span = regions[0][0][0][0][0]

    self.assertEqual(
      span.get_style(styles.StyleProperties.TextEmphasis).position,
      styles.TextEmphasisType.Position.outside
    )

  def test_direction_special_semantics(self):
    """https://github.com/sandflow/ttconv/issues/400"""
    xml_str = """<?xml version="1.0" encoding="utf-8"?>
<tt xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling" xml:lang="en">
<head>
<layout>
<region xml:id="rl" tts:writingMode="rl"/>
<region xml:id="lr" tts:writingMode="rl" tts:direction="ltr"/>
</layout>
</head>
<body>
<div>
<p region="rl" begin="0s" end="1s">tts:direction should be "rtl".</p>
<p region="lr" begin="0s" end="1s">tts:direction should be "ltr".</p>
</div>
</body>
</tt>"""

    tree = et.ElementTree(et.fromstring(xml_str))
    doc = imsc_reader.to_model(tree)

    p1 = (ISD.from_model(doc, 0).get_region("rl"))[0][0][0][0]

    self.assertEqual(
      p1.get_style(styles.StyleProperties.Direction),
      styles.DirectionType.rtl
    )

    p2 = (ISD.from_model(doc, 0).get_region("lr"))[0][0][0][0]

    self.assertEqual(
      p2.get_style(styles.StyleProperties.Direction),
      styles.DirectionType.ltr
    )

class ContentDocument1Test(unittest.TestCase):

  """
    <region xml:id="r1"/>

    <body region="r1">
      <div>
        <p begin="1s" end="3s">
          <span>
          hello
          </span>
          <span begin="1s">
          bye
          </span>
        </p>
      </div>
    </body>
  """

  def setUp(self):
    self.doc = model.ContentDocument()

    r1 = model.Region("r1", self.doc)
    r1.set_style(styles.StyleProperties.ShowBackground, styles.ShowBackgroundType.whenActive)
    self.doc.put_region(r1)

    b = model.Body(self.doc)
    b.set_region(r1)
    self.doc.set_body(b)

    div1 = model.Div(self.doc)
    b.push_child(div1)

    p1 = model.P(self.doc)
    p1.set_begin(1)
    p1.set_end(3)
    div1.push_child(p1)

    span1 = model.Span(self.doc)
    span1.push_child(model.Text(self.doc, "hello"))
    p1.push_child(span1)

    span2 = model.Span(self.doc)
    span2.set_begin(1)
    span2.push_child(model.Text(self.doc, "bye"))
    p1.push_child(span2)

  def test_sig_times(self):
  
    self.assertSequenceEqual(ISD.significant_times(self.doc), sorted((0, 1, 2, 3)))

  def test_isd_0(self):

    isd = ISD.from_model(self.doc, 0)

    self.assertEqual(len(isd), 0)

  def test_isd_1(self):

    isd = ISD.from_model(self.doc, 1)

    self.assertEqual(len(isd), 1)

    p = list(isd.iter_regions())[0][0][0][0]

    self.assertEqual(len(p), 1)

    self.assertEqual(p[0][0].get_text(), "hello")

  def test_isd_2(self):

    isd = ISD.from_model(self.doc, 2)

    self.assertEqual(len(isd), 1)

    p = list(isd.iter_regions())[0][0][0][0]

    self.assertEqual(len(p), 2)

    self.assertEqual(p[0][0].get_text(), "hello")

    self.assertEqual(p[1][0].get_text(), "bye")

  def test_isd_3(self):

    isd = ISD.from_model(self.doc, 0)

    self.assertEqual(len(isd), 0)


class DefaultRegion(unittest.TestCase):

  def test_default_region(self):

    doc = model.ContentDocument()

    b = model.Body(doc)
    doc.set_body(b)

    div1 = model.Div(doc)
    b.push_child(div1)

    p1 = model.P(doc)
    div1.push_child(p1)

    span1 = model.Span(doc)
    span1.push_child(model.Text(doc, "hello"))
    p1.push_child(span1)

    isd = ISD.from_model(doc, 0)

    self.assertEqual(len(isd), 1)

    regions = list(isd.iter_regions())

    self.assertEqual(regions[0].get_id(), ISD.DEFAULT_REGION_ID)

    p = regions[0][0][0][0]

    self.assertEqual(len(p), 1)

    self.assertEqual(p[0][0].get_text(), "hello")

if __name__ == '__main__':
  unittest.main()


class ISDReferenceFiles:
  input_dir = os.path.join("src", "test", "resources", "ttml", "imsc-tests")
  output_dir = os.path.join("src", "test", "resources", "isd-references")

  if not os.path.exists(output_dir):
    os.makedirs(output_dir)

  @staticmethod
  def walk(func: typing.Callable[[str, str], None]) -> None:
    in_files = glob.glob(os.path.join(ISDReferenceFiles.input_dir, "**", "*.ttml"), recursive=True)
    for in_file in in_files:
      in_rel_path = os.path.relpath(in_file, ISDReferenceFiles.input_dir)
      out_fn = os.path.splitext(in_rel_path)[0].replace(os.sep, "_") + ".txt"
      out_path = os.path.join(ISDReferenceFiles.output_dir, out_fn)
      func(in_file, out_path)

  @staticmethod
  def generate_reference_file(ttml_path) -> str:
    tree = et.parse(ttml_path)
    doc = imsc_reader.to_model(tree)
    sig_times = ISD.significant_times(doc)

    output_lines = []
    for time in sig_times:
      isd = ISD.from_model(doc, time)
      output_lines.append(f"{float(time)} {'+' if any(len(r) > 0 for r in isd.iter_regions()) else '-'}")
    return "\n".join(output_lines) + "\n"


class TestISDReferences(unittest.TestCase):

  def test_isd_references(self):
    def check_isd_reference(ttml_path: str, reference_path: str):
      with self.subTest(ttml_file=ttml_path):

        if not os.path.exists(reference_path):
           self.fail(f"Reference file {reference_path} does not exist for {ttml_path}")

        generated_content = ISDReferenceFiles.generate_reference_file(ttml_path)

        with open(reference_path, "r", encoding="utf-8") as f:
          reference_content = f.read()

        self.assertEqual(generated_content.replace("\r\n", "\n"), reference_content.replace("\r\n", "\n"), f"Content mismatch for {ttml_path}")

    ISDReferenceFiles.walk(check_isd_reference)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Process IMSC non-style attributes'''

import re
import logging
import math
from fractions import Fraction
import typing
from dataclasses import dataclass
from enum import Enum
from ttconv.time_code import SmpteTimeCode
import ttconv.model as model
import ttconv.imsc.utils as utils
import ttconv.imsc.namespaces as ns
from ttconv.time_code import ClockTime

LOGGER = logging.getLogger(__name__)

class XMLIDAttribute:
  '''xml:id attribute
  '''

  qn = f'{{{ns.XML}}}id'

  @staticmethod
  def extract(ttml_element):
    return ttml_element.attrib.get(XMLIDAttribute.qn)

  @staticmethod
  def set(xml_element, model_value):
    xml_element.set(XMLIDAttribute.qn, model_value)

class XMLLangAttribute:
  '''xml:lang attribute
  '''

  qn = f'{{{ns.XML}}}lang'

  @staticmethod
  def extract(ttml_element):
    return ttml_element.attrib.get(XMLLangAttribute.qn)

  @staticmethod
  def set(ttml_element, lang):
    ttml_element.set(XMLLangAttribute.qn, lang)

class XMLSpaceAttribute:
  '''xml:space attribute
  '''

  qn = f'{{{ns.XML}}}space'

  @staticmethod
  def extract(ttml_element):

    value = ttml_element.attrib.get(XMLSpaceAttribute.qn)

    r = None

    if value is not None:

      try: 
        r = model.WhiteSpaceHandling(value)
      except ValueError:
        LOGGER.error("Bad xml:space value (%s)", value)
    
    return r

  @staticmethod
  def set(ttml_element, xml_space: model.WhiteSpaceHandling):
    ttml_element.set(XMLSpaceAttribute.qn, xml_space.value)


class RegionAttribute:
  '''TTML region attribute'''

  qn = "region"

  @staticmethod
  def extract(ttml_element) -> typing.Optional[str]:
    return ttml_element.attrib.get(RegionAttribute.qn)

  @staticmethod
  def set(ttml_element, region_id: str):
    ttml_element.set(RegionAttribute.qn, region_id)


class ContentProfilesAttribute:
  '''ttp:contentProfiles attribute
  '''

  qn = f"{{{ns.TTP}}}contentProfiles"

  _CONTENT_PROFILES_RE = re.compile(r"\S+(?:\s+\S+)*")

  @staticmethod
  def extract(ttml_element) -> typing.Optional[typing.Set[str]]:

    cr = ttml_element.attrib.get(ContentProfilesAttribute.qn)

    if cr is not None:

      m = ContentProfilesAttribute._CONTENT_PROFILES_RE.match(cr)

      if m is not None:

        return set(re.findall(r'\S+', cr))

      LOGGER.error("ttp:contentProfiles invalid syntax")

    # default value in TTML

    return None

  @staticmethod
  def set(ttml_element, res):
    if res is not None:
      ttml_element.set(ContentProfilesAttribute.qn, " ".join(res))

class CellResolutionAttribute:
  '''ttp:cellResolution attribute
  '''

  qn = f"{{{ns.TTP}}}cellResolution"

  _CELL_RESOLUTION_RE = re.compile(r"(\d+) (\d+)")

  @staticmethod
  def extract(ttml_element) -> model.CellResolutionType:

    cr = ttml_element.attrib.get(CellResolutionAttribute.qn)

    if cr is not None:

      m = CellResolutionAttribute._CELL_RESOLUTION_RE.match(cr)

      if m is not None:

        return model.CellResolutionType(columns=int(m.group(1)), rows=int(m.group(2)))

      LOGGER.error("ttp:cellResolution invalid syntax")

    # default value in TTML

    return model.CellResolutionType(rows=15, columns=32)

  @staticmethod
  def set(ttml_element, res):
    ttml_element.set(CellResolutionAttribute.qn, f"{res.columns} {res.rows}")

class ExtentAttribute:
  '''ttp:extent attribute on \\<tt\\>
  '''

  qn = f"{{{ns.TTS}}}extent"

  @staticmethod
  def extract(ttml_element) -> typing.Optional[model.PixelResolutionType]:

    extent = ttml_element.attrib.get(ExtentAttribute.qn)

    if extent is not None:

      s = extent.split(" ")

      (w, w_units) = utils.parse_length(s[0])

      (h, h_units) = utils.parse_length(s[1])

      if w_units != "px" or h_units != "px":
        LOGGER.error("ttp:extent on <tt> does not use px units")
        return None

      if not w.is_integer() or not h.is_integer():
        LOGGER.error("Pixel resolution dimensions must be integer values")

      return model.PixelResolutionType(int(w), int(h))

    return None

  @staticmethod
  def set(ttml_element, res):
    ttml_element.set(ExtentAttribute.qn, f"{res.width:g}px {res.height:g}px")

class ActiveAreaAttribute:
  '''ittp:activeArea attribute on \\<tt\\>
  '''

  qn = f"{{{ns.ITTP}}}activeArea"

  @staticmethod
  def extract(ttml_element) -> typing.Optional[model.ActiveAreaType]:

    aa = ttml_element.attrib.get(ActiveAreaAttribute.qn)

    if aa is not None:

      s = aa.split(" ")

      if len(s) != 4:
        LOGGER.error("Syntax error in ittp:activeArea on <tt>")
        return None

      (left_offset, left_offset_units) = utils.parse_length(s[0])

      (top_offset, top_offset_units) = utils.parse_length(s[1])

      (w, w_units) = utils.parse_length(s[2])

      (h, h_units) = utils.parse_length(s[3])

      if w_units != "%" or h_units != "%" or left_offset_units != "%" or top_offset_units != "%":
        LOGGER.error("ittp:activeArea on <tt> must use % units")
        return None

      return model.ActiveAreaType(
        left_offset / 100,
        top_offset / 100,
        w / 100,
        h / 100
        )

    return None

  @staticmethod
  def set(ttml_element, active_area):
    ttml_element.set(
      ActiveAreaAttribute.qn, 
      f"{active_area.left_offset * 100:g}% "
      f"{active_area.top_offset * 100:g}% "
      f"{active_area.width * 100:g}% "
      f"{active_area.height * 100:g}%"
    )

class TickRateAttribute:
  '''ttp:tickRate attribute
  '''

  qn = f"{{{ns.TTP}}}tickRate"

  _TICK_RATE_RE = re.compile(r"(\d+)")

  @staticmethod
  def extract(ttml_element) -> int:

    tr = ttml_element.attrib.get(TickRateAttribute.qn)

    if tr is not None:

      m = TickRateAttribute._TICK_RATE_RE.match(tr)

      if m is not None:

        return int(m.group(1))

      LOGGER.error("ttp:tickRate invalid syntax")

    # default value

    return 1

class AspectRatioAttribute:
  '''ittp:aspectRatio attribute
  '''

  qn = f"{{{ns.ITTP}}}aspectRatio"

  _re = re.compile(r"(\d+) (\d+)")

  @staticmethod
  def extract(ttml_element) -> typing.Optional[Fraction]:

    ar_raw = ttml_element.attrib.get(AspectRatioAttribute.qn)

    if ar_raw is None:
      return None

    m = AspectRatioAttribute._re.match(ar_raw)

    if m is None:
      LOGGER.error("ittp:aspectRatio invalid syntax")
      return None

    try:

      return Fraction(int(m.group(1)), int(m.group(2)))

    except ZeroDivisionError:

      LOGGER.error("ittp:aspectRatio denominator is 0")
    
    return None

class DisplayAspectRatioAttribute:
  '''ttp:displayAspectRatio attribute
  '''

  qn = f"{{{ns.TTP}}}displayAspectRatio"

  _re = re.compile(r"(\d+) (\d+)")

  @staticmethod
  def extract(ttml_element) -> typing.Optional[Fraction]:

    ar_raw = ttml_element.attrib.get(DisplayAspectRatioAttribute.qn)

    if ar_raw is None:
      return None

    m = DisplayAspectRatioAttribute._re.match(ar_raw)

    if m is None:
      LOGGER.error("ttp:displayAspectRatio invalid syntax")
      return None

    try:

      return Fraction(int(m.group(1)), int(m.group(2)))

    except ZeroDivisionError:

      LOGGER.error("ttp:displayAspectRatio denominator is 0")
    
    return None

  @staticmethod
  def set(ttml_element, display_aspect_ratio: Fraction):
    ttml_element.set(
      DisplayAspectRatioAttribute.qn, 
      f"{display_aspect_ratio.numerator:g} {display_aspect_ratio.denominator:g}"
    )

class FrameRateAttribute:
  '''ttp:frameRate and ttp:frameRateMultiplier attribute
  '''

  frame_rate_qn = f"{{{ns.TTP}}}frameRate"

  frame_rate_multiplier_qn = f"{{{ns.TTP}}}frameRateMultiplier"

  _FRAME_RATE_RE = re.compile(r"(\d+)")

  _FRAME_RATE_MULT_RE = re.compile(r"(\d+) (\d+)")

  @staticmethod
  def extract(ttml_element) -> Fraction:

    # process ttp:frameRate

    fr = Fraction(30, 1)

    fr_raw = ttml_element.attrib.get(FrameRateAttribute.frame_rate_qn)

    if fr_raw is not None:

      m = FrameRateAttribute._FRAME_RATE_RE.match(fr_raw)

      if m is not None:

        fr = Fraction(m.group(1))

      else:

        LOGGER.error("ttp:frameRate invalid syntax")

    # process ttp:frameRateMultiplier
    
    frm = Fraction(1, 1)

    frm_raw = ttml_element.attrib.get(FrameRateAttribute.frame_rate_multiplier_qn)

    if frm_raw is not None:

      m = FrameRateAttribute._FRAME_RATE_MULT_RE.match(frm_raw)

      if m is not None:

        frm = Fraction(int(m.group(1)), int(m.group(2)))

      else:

        LOGGER.error("ttp:frameRateMultiplier invalid syntax")

    return fr * frm

  @staticmethod
  def set(ttml_element, frame_rate: Fraction):
    rounded_fps = round(frame_rate)

    ttml_element.set(
      FrameRateAttribute.frame_rate_qn, 
      str(rounded_fps)
    )

    fps_multiplier = frame_rate / rounded_fps

    if fps_multiplier != 1:

      ttml_element.set(
        FrameRateAttribute.frame_rate_multiplier_qn, 
        f"{fps_multiplier.numerator:g} {fps_multiplier.denominator:g}"
      )

class TimeBase(Enum):
  clock = "clock"
  media = "media"
  smpte = "smpte"

class DropMode(Enum):
  nonDrop = "nonDrop"
  dropNTSC = "dropNTSC"
  dropPAL = "dropPAL"

@dataclass
class TemporalAttributeParsingContext:
  frame_rate: Fraction = Fraction(30, 1)
  tick_rate: int = 1
  time_base: TimeBase = TimeBase.media
  drop_mode: DropMode = DropMode.nonDrop

class TimeExpressionSyntaxEnum(Enum):
  """IMSC time expression configuration values"""
  frames = "frames"
  clock_time = "clock_time"
  clock_time_with_frames = "clock_time_with_frames"

@dataclass
class TemporalAttributeWritingContext:
  frame_rate: typing.Optional[Fraction] = None
  time_expression_syntax: TimeExpressionSyntaxEnum = TimeExpressionSyntaxEnum.clock_time

def to_time_format(context: TemporalAttributeWritingContext, time: Fraction) -> str:
  if context.time_expression_syntax is TimeExpressionSyntaxEnum.clock_time or context.frame_rate is None:
    return str(ClockTime.from_seconds(time))

  if context.time_expression_syntax is TimeExpressionSyntaxEnum.frames:
    return f"{math.ceil(time * context.frame_rate)}f"

  return f"{SmpteTimeCode.from_seconds(time, context.frame_rate)}"

_CLOCK_TIME_FRACTION_RE = re.compile(r"^(\d{2,}):(\d\d):(\d\d(?:\.\d+)?)$")
_CLOCK_TIME_FRAMES_RE = re.compile(r"^(\d{2,}):(\d\d):(\d\d):(\d{2,})$")
_OFFSET_FRAME_RE = re.compile(r"^(\d+(?:\.\d+)?)f")
_OFFSET_TICK_RE = re.compile(r"^(\d+(?:\.\d+)?)t$")
_OFFSET_MS_RE = re.compile(r"^(\d+(?:\.\d+)?)ms$")
_OFFSET_S_RE = re.compile(r"^(\d+(?:\.\d+)?)s$")
_OFFSET_H_RE = re.compile(r"^(\d+(?:\.\d+)?)h$")
_OFFSET_M_RE = re.compile(r"^(\d+(?:\.\d+)?)m$")


def parse_time_expression(ctx: TemporalAttributeParsingContext, time_expr: str, strict: bool = True) -> Fraction:
  '''Parse a TTML time expression in a fractional number in seconds
  '''

  m = _OFFSET_FRAME_RE.match(time_expr)

  if m and ctx.frame_rate is not None:
    return Fraction(m.group(1)) / ctx.frame_rate

  m = _OFFSET_TICK_RE.match(time_expr)

  if m and ctx.tick_rate is not None:
    return Fraction(m.group(1)) / ctx.tick_rate

  m = _OFFSET_MS_RE.match(time_expr)

  if m:
    return Fraction(m.group(1)) / 1000

  m = _OFFSET_S_RE.match(time_expr)

  if m:
    return Fraction(m.group(1))

  m = _OFFSET_M_RE.match(time_expr)

  if m:
    return Fraction(m.group(1)) * 60

  m = _OFFSET_H_RE.match(time_expr)

  if m:
    return Fraction(m.group(1)) * 3600

  m = _CLOCK_TIME_FRACTION_RE.match(time_expr)

  if m:
    return Fraction(m.group(1)) * 3600 + \
            Fraction(m.group(2)) * 60 + \
            Fraction(m.group(3))

  m = _CLOCK_TIME_FRAMES_RE.match(time_expr)

  if m and ctx.frame_rate is not None:
    frames = int(m.group(4)) if m.group(4) else 0

    if frames >= ctx.frame_rate:
      if strict:
        raise ValueError("Frame count exceeds frame rate")
      else:
        LOGGER.error("Frame count %s exceeds frame rate %s, rounding to frame rate minus 1", frames, ctx.frame_rate)
        frames = round(ctx.frame_rate) - 1

    hh = int(m.group(1))
    mm = int(m.group(2))
    ss = int(m.group(3))

    if ctx.time_base is TimeBase.smpte:
      tc = SmpteTimeCode(hh, mm, ss, frames, ctx.frame_rate, ctx.drop_mode != DropMode.nonDrop)
      return tc.to_temporal_offset()
    else:
      return Fraction(hh * 3600 + mm * 60 + ss) + Fraction(frames) / ctx.frame_rate

  raise ValueError("Syntax error")

class BeginAttribute:
  '''begin attribute
  '''
  qn = "begin"

  @staticmethod
  def extract(context: TemporalAttributeParsingContext, xml_element) -> typing.Optional[Fraction]:

    # read begin attribute

    begin_raw = xml_element.attrib.get(BeginAttribute.qn)

    try:
      return parse_time_expression(context, begin_raw, False) if begin_raw is not None else None
    except ValueError:
      LOGGER.error("Bad begin attribute value: %s", begin_raw)
      return None

  @staticmethod
  def set(context: TemporalAttributeWritingContext, ttml_element, begin:Fraction):
    value = to_time_format(context, begin)
    ttml_element.set(BeginAttribute.qn, value)

class EndAttribute:
  '''end attributes
  '''
  qn = "end"

  @staticmethod
  def extract(context: TemporalAttributeParsingContext, xml_element) -> typing.Optional[Fraction]:

    # read end attribute

    end_raw = xml_element.attrib.get(EndAttribute.qn)

    try:
      return parse_time_expression(context, end_raw, False) if end_raw is not None else None
    except ValueError:
      LOGGER.error("Bad end attribute value: %s", end_raw)
      return None

  @staticmethod
  def set(context: TemporalAttributeWritingContext, ttml_element, end:Fraction):
    value = to_time_format(context, end)
    ttml_element.set(EndAttribute.qn, value)

class DurAttribute:
  '''dur attributes
  '''
  qn = "dur"

  @staticmethod
  def extract(context: TemporalAttributeParsingContext, xml_element) -> typing.Optional[Fraction]:

    dur_raw = xml_element.attrib.get(DurAttribute.qn)

    try:
      return parse_time_expression(context, dur_raw, False) if dur_raw is not None else None
    except ValueError:
      LOGGER.error("Bad dur attribute value: %s", dur_raw)
      return None

  @staticmethod
  def set(ttml_element, dur):
    raise NotImplementedError

class TimeContainer(Enum):
  par = "par"
  seq = "seq"

  def is_seq(self) -> bool:
    return self == TimeContainer.seq

  def is_par(self) -> bool:
    return self == TimeContainer.par

class TimeContainerAttribute:
  '''timeContainer attributes
  '''

  qn = "timeContainer"

  @staticmethod
  def extract(xml_elem) -> TimeContainer:

    time_container_raw = xml_elem.attrib.get(TimeContainerAttribute.qn)

    try:

      return TimeContainer(time_container_raw) if time_container_raw is not None else TimeContainer.par

    except ValueError:

      LOGGER.error("bad timeContainer value")

      return TimeContainer.par


class StyleAttribute:
  '''style attribute
  '''
  qn = "style"

  @staticmethod
  def extract(xml_element) -> typing.List[str]:

    raw_value = xml_element.attrib.get(StyleAttribute.qn)

    return raw_value.split(" ") if raw_value is not None else []

class TimeBaseAttribute:
  '''ttp:timeBase attribute
  '''

  qn = f"{{{ns.TTP}}}timeBase"

  @staticmethod
  def extract(ttml_element) -> TimeBase:

    cr = ttml_element.attrib.get(TimeBaseAttribute.qn)

    if cr is None:
      return TimeBase.media

    try:
      tb = TimeBase[cr]
    except KeyError:
      LOGGER.error(f"Bad ttp:timeBase value '{cr}', using 'media' instead")
      return TimeBase.media

    if tb is TimeBase.clock:
      raise ValueError("Clock timebase not supported")

    return tb

  # We do not support writing ttp:timeBase since all model values are always in media timebase

class DropModeAttribute:
  '''ttp:dropMode attribute
  '''

  qn = f"{{{ns.TTP}}}dropMode"

  @staticmethod
  def extract(ttml_element) -> DropMode:

    cr = ttml_element.attrib.get(DropModeAttribute.qn)

    if cr is None:
      return DropMode.nonDrop

    try:
      dm = DropMode[cr]
    except KeyError:
      LOGGER.error(f"Bad ttp:dropMode value '{cr}', using 'nonDrop' instead")
      return DropMode.nonDrop

    if dm is DropMode.dropPAL:
      raise ValueError("PAL drop frame timecode not supported")

    return dm

  # We do not support writing ttp:dropMode since all model values are always in media timebase


#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Filter that remove animations"""

import logging
from typing import Dict, List, Type

from ttconv.model import ContentDocument, ContentElement
from ttconv.style_properties import StyleProperty

class RemoveAnimationFilter:
  """Filter that remove animations"""

  def __init__(self) -> None:
    self._has_removed_animations = False

  def has_removed_animations(self) -> bool:
    return self._has_removed_animations

  def process_element(self, element: ContentElement, recursive = True):
    """Removes animations from content elements"""

    for step in element.iter_animation_steps():
      element.remove_animation_step(step)
      self._has_removed_animations = True

    if recursive:
      for child in element:
        self.process_element(child)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''IMSC style properties'''

import logging
import re
import typing
import ttconv.style_properties as styles

LOGGER = logging.getLogger(__name__)

_LENGTH_RE = re.compile(r"^((?:\+|\-)?\d*(?:\.\d+)?)(px|em|c|%|rh|rw)$")

def parse_length(attr_value: str) -> typing.Tuple[float, str]:
  '''Parses the TTML length in `attr_value` into a (length, units) tuple'''

  m = _LENGTH_RE.match(attr_value)

  if m:

    return (float(m.group(1)), m.group(2))

  raise ValueError("Bad length syntax")


_FAMILIES_ESCAPED_CHAR = re.compile(r"\\(.)")
_SINGLE_QUOTE_PATTERN = "(?:'(?P<single_quote>(.+?)(?<!\\\\))')"
_DOUBLE_QUOTE_PATTERN = "(?:\"(?P<double_quote>(.+?)(?<!\\\\))\")"
_NO_QUOTE_PATTERN = "(?P<no_quote>(?:\\\\.|[^'\", ])(?:\\\\.|[^'\",])+)"

_FONT_FAMILY_PATTERN = re.compile(
  "|".join(
    (
      _SINGLE_QUOTE_PATTERN,
      _DOUBLE_QUOTE_PATTERN,
      _NO_QUOTE_PATTERN
    )
  )
)

def parse_font_families(attr_value: str) -> typing.List[str]:
  '''Parses th TTML \\<font-family\\> value in `attr_value` into a list of font families'''

  rslt = []

  for m in _FONT_FAMILY_PATTERN.finditer(attr_value):

    is_quoted = m.lastgroup in ("single_quote", "double_quote")

    escaped_family = _FAMILIES_ESCAPED_CHAR.sub(r"\1", m.group(m.lastgroup))

    if not is_quoted and escaped_family in styles.GenericFontFamilyType.__members__:
      rslt.append(styles.GenericFontFamilyType(escaped_family))
    else:
      rslt.append(escaped_family)
    

  if len(rslt) == 0:
    raise ValueError("Bad syntax")

  return rslt

def serialize_font_family(font_family: typing.Tuple[typing.Union[str, styles.GenericFontFamilyType], ...]):
  '''Serialize model FontFamily to tts:fontFamily
  '''

  def _serialize_one_family(family):
    if isinstance(family, styles.GenericFontFamilyType):
      return family.value
    
    return '"' + family.replace('"', r'\"') + '"'

  return ", ".join(map(_serialize_one_family, font_family))

def parse_position(attr_value: str) -> typing.Tuple[str, styles.LengthType, str, styles.LengthType]:
  '''Parse a TTML \\<position\\> value into offsets from a horizontal and vertical edge
  '''

  length_50pct = styles.LengthType(value=50, units=styles.LengthType.Units.pct)
  length_0pct = styles.LengthType(value=0, units=styles.LengthType.Units.pct)

  h_edges = {"left", "right"}
  v_edges = {"top", "bottom"}

  h_edge: typing.Optional[str] = None
  h_offset: typing.Optional[styles.LengthType] = None
  v_edge: typing.Optional[str] = None
  v_offset: typing.Optional[styles.LengthType] = None

  items = attr_value.split()

  if len(items) in (1, 2):

    # begin processing 1 and 2 components

    while len(items) > 0:

      cur_item = items.pop(0)

      if cur_item in h_edges:

        h_edge = cur_item
        h_offset = length_0pct

      elif cur_item in v_edges:

        v_edge = cur_item
        v_offset = length_0pct

      elif cur_item == "center":

        if h_edge is None:

          h_edge = "left"
          h_offset = length_50pct

        elif v_edge is None:

          v_edge = "top"
          v_offset = length_50pct

      else:

        (value, units) = parse_length(cur_item)

        if h_edge is None:

          h_edge = "left"
          h_offset = styles.LengthType(value, styles.LengthType.Units(units))

        elif v_edge is None:

          v_edge = "top"
          v_offset = styles.LengthType(value, styles.LengthType.Units(units))
    
    # end processing 1 and 2 components

  else:

    # begin processing 3 and 4 components

    while len(items) > 0:

      cur_item = items.pop(0)

      if cur_item in h_edges:

        h_edge = cur_item

        if v_edge is not None and v_offset is None:
          v_offset = length_0pct

      elif cur_item in v_edges:

        v_edge = cur_item

        if h_edge is not None and h_offset is None:
          h_offset = length_0pct

      elif cur_item == "center":

        pass

      else:

        (value, units) = parse_length(cur_item)

        if h_edge is not None and h_offset is None:

          h_offset = styles.LengthType(value, styles.LengthType.Units(units))

        if v_edge is not None and v_offset is None:

          v_offset = styles.LengthType(value, styles.LengthType.Units(units))
    
    # end processing 3 and 4 components

  # fill-in missing components, if any

  if h_offset is None:

    if h_edge is None:
      h_edge = "left"
      h_offset = length_50pct
    else:
      h_offset = length_0pct

  if v_offset is None:

    if v_edge is None:
      v_edge = "top"
      v_offset = length_50pct
    else:
      v_offset = length_0pct


  return (h_edge, h_offset, v_edge, v_offset)
  


#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for SCC disassembly format conversion"""

# pylint: disable=R0201,C0115,C0116,W0212

import unittest

from ttconv.scc.disassembly import get_color_disassembly, get_font_style_disassembly, get_text_decoration_disassembly
from ttconv.scc.reader import to_disassembly
from ttconv.style_properties import NamedColors, ColorType, FontStyleType, TextDecorationType


class SccReaderTest(unittest.TestCase):

  def test_scc_disassembly_color(self):
    disassembly_colors = {
      NamedColors.white: "Wh",
      NamedColors.green: "Gr",
      NamedColors.blue: "Bl",
      NamedColors.cyan: "Cy",
      NamedColors.aqua: "Cy",
      NamedColors.red: "R",
      NamedColors.yellow: "Y",
      NamedColors.magenta: "Ma",
      NamedColors.fuchsia: "Ma",
      NamedColors.black: "Bk",
      NamedColors.transparent: "T",
      NamedColors.silver: "",
      NamedColors.gray: "",
      NamedColors.maroon: "",
      NamedColors.purple: "",
      NamedColors.lime: "",
      NamedColors.olive: "",
      NamedColors.navy: "",
      NamedColors.teal: ""
    }

    for (color, code) in disassembly_colors.items():
      self.assertEqual(code, get_color_disassembly(color.value))

    self.assertEqual("BlS", get_color_disassembly(ColorType((0, 0, 255, 136))))
    self.assertEqual("Bl", get_color_disassembly(ColorType((0, 0, 255, 128))))
    self.assertEqual("Bl", get_color_disassembly(ColorType((0, 0, 255, 145))))
    self.assertEqual("T", get_color_disassembly(ColorType((0, 0, 255, 0))))

  def test_scc_disassembly_font_style(self):
    disassembly_font_style = {
      FontStyleType.normal: "",
      FontStyleType.oblique: "",
      FontStyleType.italic: "I",
    }

    for (font_style, code) in disassembly_font_style.items():
      self.assertEqual(code, get_font_style_disassembly(font_style))

  def test_scc_disassembly_text_decoration(self):
    disassembly_text_decoration = {
      TextDecorationType(underline=True): "U",
      TextDecorationType(underline=False): "",
      TextDecorationType(underline=None): "",
      TextDecorationType(overline=True): "",
      TextDecorationType(line_through=True): "",
      None: ""
    }

    for (text_decoration, code) in disassembly_text_decoration.items():
      self.assertEqual(code, get_text_decoration_disassembly(text_decoration))

  def test_scc_disassembly(self):
    scc_content = """Scenarist_SCC V1.0

00:00:00:22	9425 9425 94ad 94ad 9470 9470 4c6f 7265 6d20 6970 7375 6d20 646f 6c6f 7220 7369 7420 616d 6574 2c80

00:00:02:23	9425 9425 94ad 94ad 9673 9673 636f 6e73 6563 7465 7475 7220 6164 6970 6973 6369 6e67 2065 6c69 742e

00:00:20;06	9426 9426 94ad 94ad 9470 9470 5065 6c6c 656e 7465 7371 7565 2069 6e74 6572 6475 6d20 6c61 6369 6e69 6120 736f 6c6c 6963 6974 7564 696e 2e80

00:00:21;24	9426 9426 94ad 94ad 9470 9470 496e 7465 6765 7220 6c75 6374 7573 2065 7420 6c69 6775 6c61 2061 6320 7361 6769 7474 6973 2e80

00:00:49;03	94a7 94ad 9470 5574 2061 7420 6469 616d 2073 6974 2061 6d65 7420 6e75 6c6c 6120 6672 696e 6769 6c6c 6180

00:00:50;23	94a7 94ad 9470 7665 7374 6962 756c 756d 206e 6563 2076 6974 6165 206e 6973 692e

00:02:53:14	9429 9429 94d2 94d2 4c6f 7265 6d20 6970 7375 6d20 646f 6c6f 7220 7369 7420 616d 6574 2c80 94f2 94f2 636f 6e73 6563 7465 7475 7220 6164 6970 6973 6369 6e67 2065 6c69 742e

00:02:56:00	9429 9429 94d2 94d2 5065 6c6c 656e 7465 7371 7565 2069 6e74 6572 6475 6d20 6c61 6369 6e69 6120 736f 6c6c 6963 6974 7564 696e 2e80

00:02:56:25	9429 9429 94f2 94f2 496e 7465 6765 7220 6c75 6374 7573 2065 7420 6c69 6775 6c61 2061 6320 7361 6769 7474 6973 2e80

01:02:53:14	94ae 94ae 9420 9420 947a 947a 97a2 97a2 a820 68ef f26e 2068 ef6e 6be9 6e67 2029 942c 942c 8080 8080 942f 942f

01:02:55:14	942c 942c

01:03:27:29	94ae 94ae 9420 9420 94f2 94f2 c845 d92c 2054 c845 91b0 45ae 942c 942c 8080 8080 942f 942f

01:11:31:01	9420 9420 9452 9452 97a1 97a1 54e5 73f4 2080 9132 2043 6170 f4e9 ef6e 2080 94f2 94f2 97a1 97a1 54e5 73f4 2080 91ae 91ae f4e5 73f4 9120 9120 2043 6170 f4e9 ef6e 7380 942c 942c 942f 942f

01:11:33:14	942c 942c
"""

    expected = """00:00:00:22	{RU2}{RU2}{CR}{CR}{1500}{1500}Lorem ipsum dolor sit amet,
00:00:02:23	{RU2}{RU2}{CR}{CR}{0804}{0804}consectetur adipiscing elit.
00:00:20;06	{RU3}{RU3}{CR}{CR}{1500}{1500}Pellentesque interdum lacinia sollicitudin.
00:00:21;24	{RU3}{RU3}{CR}{CR}{1500}{1500}Integer luctus et ligula ac sagittis.
00:00:49;03	{RU4}{CR}{1500}Ut at diam sit amet nulla fringilla
00:00:50;23	{RU4}{CR}{1500}vestibulum nec vitae nisi.
00:02:53:14	{RDC}{RDC}{1404}{1404}Lorem ipsum dolor sit amet,{1504}{1504}consectetur adipiscing elit.
00:02:56:00	{RDC}{RDC}{1404}{1404}Pellentesque interdum lacinia sollicitudin.
00:02:56:25	{RDC}{RDC}{1504}{1504}Integer luctus et ligula ac sagittis.
01:02:53:14	{ENM}{ENM}{RCL}{RCL}{1520}{1520}{TO2}{TO2}( horn honking ){EDM}{EDM}{}{}{EOC}{EOC}
01:02:55:14	{EDM}{EDM}
01:03:27:29	{ENM}{ENM}{RCL}{RCL}{1504}{1504}HEY, THE®E.{EDM}{EDM}{}{}{EOC}{EOC}
01:11:31:01	{RCL}{RCL}{1404}{1404}{TO1}{TO1}Test ½ Caption {1504}{1504}{TO1}{TO1}Test {I}{I}test{Wh}{Wh} Captions{EDM}{EDM}{EOC}{EOC}
01:11:33:14	{EDM}{EDM}
"""

    disassembly = to_disassembly(scc_content)

    self.assertEqual(expected, disassembly)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""CEA-608 Standard characters"""

from __future__ import annotations

SCC_STANDARD_CHARACTERS_MAPPING = dict([
  (0x20, " "),  # Standard space
  (0x21, "!"),  # Exclamation mark
  (0x22, "\""),  # Quotation mark
  (0x23, "#"),  # Pounds (number) sign
  (0x24, "$"),  # Dollar sign
  (0x25, "%"),  # Percentage sign
  (0x26, "&"),  # Ampersand
  (0x27, "'"),  # Apostrophe
  (0x28, "("),  # Open parentheses
  (0x29, ")"),  # Close parentheses
  (0x2A, "\u00E1"),  # á Lower-case a with acute accent
  (0x2B, "+"),  # Plus sign
  (0x2C, ","),  # Comma
  (0x2D, "-"),  # Minus (hyphen) sign
  (0x2E, "."),  # Period
  (0x2F, "/"),  # Slash
  (0x30, "0"),  # Zero
  (0x31, "1"),  # One
  (0x32, "2"),  # Two
  (0x33, "3"),  # Three
  (0x34, "4"),  # Four
  (0x35, "5"),  # Five
  (0x36, "6"),  # Six
  (0x37, "7"),  # Seven
  (0x38, "8"),  # Eight
  (0x39, "9"),  # Nine
  (0x3A, ":"),  # Colon
  (0x3B, ";"),  # Semi-colon
  (0x3C, "<"),  # Less than sign
  (0x3D, "="),  # Equal sign
  (0x3E, ">"),  # Greater than sign
  (0x3F, "?"),  # Question mark
  (0x40, "@"),  # At sign
  (0x41, "A"),  # Upper-case A
  (0x42, "B"),  # Upper-case B
  (0x43, "C"),  # Upper-case C
  (0x44, "D"),  # Upper-case D
  (0x45, "E"),  # Upper-case E
  (0x46, "F"),  # Upper-case F
  (0x47, "G"),  # Upper-case G
  (0x48, "H"),  # Upper-case H
  (0x49, "I"),  # Upper-case I
  (0x4A, "J"),  # Upper-case J
  (0x4B, "K"),  # Upper-case K
  (0x4C, "L"),  # Upper-case L
  (0x4D, "M"),  # Upper-case M
  (0x4E, "N"),  # Upper-case N
  (0x4F, "O"),  # Upper-case O
  (0x50, "P"),  # Upper-case P
  (0x51, "Q"),  # Upper-case Q
  (0x52, "R"),  # Upper-case R
  (0x53, "S"),  # Upper-case S
  (0x54, "T"),  # Upper-case T
  (0x55, "U"),  # Upper-case U
  (0x56, "V"),  # Upper-case V
  (0x57, "W"),  # Upper-case W
  (0x58, "X"),  # Upper-case X
  (0x59, "Y"),  # Upper-case Y
  (0x5A, "Z"),  # Upper-case Z
  (0x5B, "["),  # Open bracket
  (0x5C, "\u00E9"),  # é Lower-case e with acute accent
  (0x5D, "]"),  # Close bracket
  (0x5E, "\u00ED"),  # í Lower-case i with acute accent
  (0x5F, "\u00F3"),  # ó Lower-case o with acute accent
  (0x60, "\u00FA"),  # ú Lower-case u with acute accent
  (0x61, "a"),  # Lower-case a
  (0x62, "b"),  # Lower-case b
  (0x63, "c"),  # Lower-case c
  (0x64, "d"),  # Lower-case d
  (0x65, "e"),  # Lower-case e
  (0x66, "f"),  # Lower-case f
  (0x67, "g"),  # Lower-case g
  (0x68, "h"),  # Lower-case h
  (0x69, "i"),  # Lower-case i
  (0x6A, "j"),  # Lower-case j
  (0x6B, "k"),  # Lower-case k
  (0x6C, "l"),  # Lower-case l
  (0x6D, "m"),  # Lower-case m
  (0x6E, "n"),  # Lower-case n
  (0x6F, "o"),  # Lower-case o
  (0x70, "p"),  # Lower-case p
  (0x71, "q"),  # Lower-case q
  (0x72, "r"),  # Lower-case r
  (0x73, "s"),  # Lower-case s
  (0x74, "t"),  # Lower-case t
  (0x75, "u"),  # Lower-case u
  (0x76, "v"),  # Lower-case v
  (0x77, "w"),  # Lower-case w
  (0x78, "x"),  # Lower-case x
  (0x79, "y"),  # Lower-case y
  (0x7A, "z"),  # Lower-case z
  (0x7B, "\u00E7"),  # ç Lower-case c with cedilla
  (0x7C, "\u00F7"),  # ÷ Division sign
  (0x7D, "\u00D1"),  # Ñ Upper-case N with tilde
  (0x7E, "\u00F1"),  # ñ Lower-case n with tilde
  (0x7F, "\u2588"),  # █ Solid block
])

TO_SCC_BYTES = {unicode_char: scc_byte for scc_byte, unicode_char in SCC_STANDARD_CHARACTERS_MAPPING.items()}



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Filters style properties"""

import logging
from typing import Dict, List, Type

from ttconv.model import ContentDocument, ContentElement
from ttconv.style_properties import StyleProperty

class SupportedStylePropertiesFilter:
  """Filter that removes unsupported style properties"""

  def __init__(self, supported_style_properties: Dict[Type[StyleProperty], List]):
    self.supported_style_properties = supported_style_properties

  def process_initial_values(self, doc: ContentDocument):
    """Removes initial values that target unsupported style properties"""
    for style_prop, value in list(doc.iter_initial_values()):

      if style_prop in self.supported_style_properties:
        supported_values = self.supported_style_properties[style_prop]

        if len(supported_values) == 0 or value in supported_values:
          continue

      doc.put_initial_value(style_prop, None)

  def process_element(self, element: ContentElement, recursive = True):
    """Removes unsupported style properties from content elements"""

    for style_prop in list(element.iter_styles()):

      if style_prop in self.supported_style_properties:
        value = element.get_style(style_prop)
        supported_values = self.supported_style_properties[style_prop]

        if len(supported_values) == 0 or value in supported_values:
          continue

      element.set_style(style_prop, None)

    if recursive:
      for child in element:
        self.process_element(child)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC caption style"""

from enum import Enum


class SccCaptionStyle(Enum):
  """SCC caption style"""
  Unknown = 0

  # SCC Roll-Up captions are composed with:
  #  - RU2 or RU3 or RU4 (to select the Roll-Up style and the number of displayed rows)
  #  - CR (to roll the displayed rows up)
  #  - PAC (to set row, indent and/or attributes)
  #  - text
  RollUp = 1

  # SCC Paint-On captions are composed with:
  #  - RDC (to select the Paint-On style)
  #  - PAC (to set row, indent and/or attributes)
  #  - text
  #  - PAC (to change row, indent and/or attributes)
  #  - text or DER (to erase the following text of the current row)
  PaintOn = 2

  # SCC Pop-On captions are composed with:
  #  - RCL (to select the Pop-On style)
  #  - ENM (to clear the memory, optional)
  #  - PAC (to set row, indent and/or attributes)
  #  - text
  #  - PAC (to change row, indent and/or attributes)
  #  - text
  #  etc.
  #  - EDM (to erase the displayed caption, optional)
  #  - EOC (to display the current caption)
  PopOn = 3

  @staticmethod
  def default():
    """Returns the default caption style"""
    return SccCaptionStyle.PopOn
